#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Post a message to slack."""
import logging

import slacker

from ..consts import SLACK_TOKEN

LOG = logging.getLogger(__name__)


def post_slack_message(message=None, channel=None, username=None, icon_emoji=None):
    """Format the message and post to the appropriate slack channel.

    Args:
        message (str): Message to post to slack
        channel (str): Desired channel. Must start with #

    """
    LOG.debug('Slack Channel: %s\nSlack Message: %s', channel, message)
    slack = slacker.Slacker(SLACK_TOKEN)
    try:
        slack.chat.post_message(channel=channel, text=message, username=username, icon_emoji=icon_emoji)
        LOG.info('Message posted to %s', channel)
    except slacker.Error:
        LOG.info("error posted message to %s", channel)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Complete Application destroyer script."""
import argparse
import logging

from boto3.exceptions import botocore

from .args import add_app, add_debug
from .consts import ENVS, LOGGING_FORMAT, REGIONS
from .dns.destroy_dns.destroy_dns import destroy_dns
from .elb.destroy_elb.destroy_elb import destroy_elb
from .exceptions import SpinnakerError
from .iam.destroy_iam.destroy_iam import destroy_iam
from .s3.destroy_s3.destroy_s3 import destroy_s3
from .securitygroup.destroy_sg.destroy_sg import destroy_sg

LOG = logging.getLogger(__name__)


def main():  # noqa
    """Attempt to fully destroy AWS Resources for a Spinnaker Application."""
    logging.basicConfig(format=LOGGING_FORMAT)

    parser = argparse.ArgumentParser(description=main.__doc__)
    add_debug(parser)
    add_app(parser)
    args = parser.parse_args()

    if args.debug == logging.DEBUG:
        logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)
    else:
        LOG.setLevel(args.debug)

    for env in ENVS:
        for region in REGIONS:
            LOG.info('DESTROY %s:%s', env, region)

            try:
                destroy_dns(app=args.app, env=env)
            except botocore.exceptions.ClientError as error:
                LOG.warning('DNS issue for %s in %s: %s', env, region, error)

            try:
                destroy_elb(app=args.app, env=env, region=region)
            except SpinnakerError:
                pass

            try:
                destroy_iam(app=args.app, env=env)
            except botocore.exceptions.ClientError as error:
                LOG.warning('IAM issue for %s in %s: %s', env, region, error)

            try:
                destroy_s3(app=args.app, env=env)
            except botocore.exceptions.ClientError as error:
                LOG.warning('S3 issue for %s in %s: %s', env, region, error)

            try:
                destroy_sg(app=args.app, env=env, region=region)
            except SpinnakerError:
                pass

            LOG.info('Destroyed %s:%s', env, region)

    LOG.info('Destruction complete.')


if __name__ == '__main__':
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Retrieve Account Credential from Gate API."""
import logging

from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


def get_env_credential(env='dev'):
    """Get Account Credential from Spinnaker for *env*.

    Args:
        env (str): Environment name to find credentials for.

    Returns:
        dict: Complete credentials for *env*::

            {
                'accountId': '123098123',
                'accountType': 'dev',
                'assumeRole': 'role/spinnakerManaged',
                'bastionEnabled': False,
                'challengeDestructiveActions': False,
                'cloudProvider': 'aws',
                'defaultKeyPair': 'dev_access',
                'discoveryEnabled': False,
                'eddaEnabled': False,
                'environment': 'dev',
                'front50Enabled': False,
                'name': 'dev',
                'primaryAccount': False,
                'provider': 'aws',
                'regions': [
                    {
                        'availabilityZones': ['us-east-1b', 'us-east-1c',
                                              'us-east-1d', 'us-east-1e'],
                        'deprecated': False,
                        'name': 'us-east-1',
                        'preferredZones':
                        ['us-east-1b', 'us-east-1c', 'us-east-1d', 'us-east-1e'
                         ]
                    }, {
                        'availabilityZones':
                        ['us-west-2a', 'us-west-2b', 'us-west-2c'],
                        'deprecated': False,
                        'name': 'us-west-2',
                        'preferredZones':
                        ['us-west-2a', 'us-west-2b', 'us-west-2c']
                    }
                ],
                'requiredGroupMembership': [],
                'sessionName': 'Spinnaker',
                'type': 'aws'
            }

    """
    uri = '/credentials/{env}'.format(env=env)
    credential_response = gate_request(uri=uri)

    assert credential_response.ok, 'Could not get credentials from Spinnaker.'

    credential = credential_response.json()
    LOG.debug('Credentials found:\n%s', credential)
    return credential



"""Base App."""
import copy
import logging

from pprint import pformat
from ..consts import LINKS
from ..exceptions import ForemastError
from ..utils import get_template, wait_for_task
from ..utils.gate import gate_request


# pylint: disable=abstract-method
class SpinnakerApp:
    """Base App."""

    def __init__(self, provider, pipeline_config=None, app=None, email=None, project=None, repo=None):
        """Class to manage and create Spinnaker applications

        Args:
            pipeline_config (dict): pipeline.json data.
            app (str): Application name.
            email (str): Email associated with application.
            project (str): Git namespace or project group
            repo (str): Repository name

        """

        self.log = logging.getLogger(__name__)

        self.appinfo = {
            'app': app,
            'email': email,
            'project': project,
            'repo': repo,
            'provider': provider
        }

        self.appname = app
        self.pipeline_config = pipeline_config

    def create(self):
        """Send a POST to spinnaker to create a new application with class variables.

                Raises:
                    AssertionError: Application creation failed.

                """

        # Retaining abstract account list for backwards compatibility
        # Refer to #366
        self.appinfo['accounts'] = ['default']
        self.log.debug('Pipeline Config\n%s', pformat(self.pipeline_config))
        self.log.debug('App info:\n%s', pformat(self.appinfo))
        jsondata = self.render_application_template()
        wait_for_task(jsondata)

        self.log.info("Successfully created %s application", self.appname)
        return jsondata

    def render_application_template(self):
        """Render application from configs.

        Returns:
            dict: Rendered application template.
        """
        self.pipeline_config['instance_links'] = self.retrieve_instance_links()
        jsondata = get_template(
            template_file='infrastructure/app_data.json.j2', appinfo=self.appinfo, pipeline_config=self.pipeline_config)
        return jsondata

    def retrieve_instance_links(self):
        """Combine default and configuration instance links.

        Returns:
            dict: Combined instance links.
        """
        instance_links = copy.copy(LINKS)
        self.log.debug('Default instance links: %s', instance_links)
        instance_links.update(self.pipeline_config['instance_links'])
        self.log.debug('Updated instance links: %s', instance_links)

        return instance_links

    def get_accounts(self):
        """Get Accounts added to Spinnaker.

        Returns:
            list: list of dicts of Spinnaker credentials matching _provider_.

        Raises:
            AssertionError: Failure getting accounts from Spinnaker.
        """
        uri = '/credentials'
        response = gate_request(uri=uri)
        assert response.ok, 'Failed to get accounts: {0}'.format(response.text)

        all_accounts = response.json()
        self.log.debug('Accounts in Spinnaker:\n%s', all_accounts)

        filtered_accounts = []
        for account in all_accounts:
            if account['type'] == self.provider:
                filtered_accounts.append(account)

        if not filtered_accounts:
            raise ForemastError('No Accounts matching {0}.'.format(self.provider))

        return filtered_accounts



"""Verify :func:`foremsat.utils.tasks.check_task` functionality."""
from unittest import mock

import pytest

from foremast.exceptions import SpinnakerTaskError, SpinnakerTaskInconclusiveError
from foremast.utils.tasks import _check_task, check_task

FAIL_MESSAGE = 'TERMINAL'
SUCCESS_MESSAGE = 'SUCCEEDED'


@mock.patch('foremast.utils.tasks._check_task')
def test_utils_retry_task(mock_check_task):
    """Validate task retries are configurable."""
    taskid = 'fake_task'
    mock_check_task.side_effect = ValueError
    with pytest.raises(SpinnakerTaskInconclusiveError):
        check_task(taskid, timeout=2, wait=1)
    assert mock_check_task.call_count == 2


@mock.patch('foremast.utils.tasks.gate_request')
def test_task_success(mock_gate_request):
    """Successful Task."""
    mock_gate_request.return_value.json.return_value = {'status': SUCCESS_MESSAGE}

    result = _check_task(taskid='')

    assert result == SUCCESS_MESSAGE


@mock.patch('foremast.utils.tasks.gate_request')
def test_task_failure(mock_gate_request):
    """Failed Task."""
    mock_gate_request.return_value.json.return_value = {
        'status': FAIL_MESSAGE,
        'execution': {
            'stages': [],
        },
    }

    with pytest.raises(SpinnakerTaskError):
        _check_task(taskid='')


@mock.patch('foremast.utils.tasks.gate_request')
def test_task_unknown(mock_gate_request):
    """Unknown Task status raises exception to keep polling."""
    mock_gate_request.get.return_value.json.return_value = {'status': ''}

    with pytest.raises(ValueError):
        _check_task(taskid='')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Get security group id."""
import logging

from tryagain import retries

from ..consts import SECURITYGROUP_REPLACEMENTS
from ..exceptions import SpinnakerSecurityGroupError
from ..utils import gate_request
from .vpc import get_vpc_id

LOG = logging.getLogger(__name__)


@retries(max_attempts=5, wait=2, exceptions=(SpinnakerSecurityGroupError))
def get_security_group_id(name='', env='', region=''):
    """Get a security group ID.

    Args:
        name (str): Security Group name to find.
        env (str): Deployment environment to search.
        region (str): AWS Region to search.

    Returns:
        str: ID of Security Group, e.g. sg-xxxx.

    Raises:
        AssertionError: Call to Gate API was not successful.
        SpinnakerSecurityGroupError: Security Group _name_ was not found for
            _env_ in _region_.

    """
    vpc_id = get_vpc_id(env, region)

    LOG.info('Find %s sg in %s [%s] in %s', name, env, region, vpc_id)

    uri = '/securityGroups/{0}/{1}/{2}?vpcId={3}'.format(env, region, name, vpc_id)
    response = gate_request(uri=uri)
    assert response.ok

    result = response.json()
    try:
        security_group_id = result['id']
    except KeyError:
        msg = 'Security group ({0}) not found'.format(name)
        raise SpinnakerSecurityGroupError(msg)

    LOG.info('Found: %s', security_group_id)
    return security_group_id


def remove_duplicate_sg(security_groups):
    """Removes duplicate Security Groups that share a same name alias

    Args:
        security_groups (list): A list of security group id to compare against SECURITYGROUP_REPLACEMENTS

    Returns:
        security_groups (list): A list of security groups with duplicate aliases removed
    """
    for each_sg, duplicate_sg_name in SECURITYGROUP_REPLACEMENTS.items():
        if each_sg in security_groups and duplicate_sg_name in security_groups:
            LOG.info('Duplicate SG found. Removing %s in favor of %s.', duplicate_sg_name, each_sg)
            security_groups.remove(duplicate_sg_name)

    return security_groups



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Ensure Security Groups get created properly"""
import json
from unittest import mock

import pytest

from foremast.exceptions import ForemastConfigurationFileError
from foremast.securitygroup import SpinnakerSecurityGroup

SAMPLE_JSON = """{"security_group": {
                    "description": "something useful",
                    "egress": "0.0.0.0/0",
                    "elb_extras": [],
                    "ingress": {
                      "coreforrest": [{
                        "start_port": 8080, "end_port": 8080,
                        "env": "dev", "protocol": "tcp"
                      }],
                      "sg_apps": [{
                        "start_port": 80, "end_port": 80, "protocol": "tcp"
                      }]
                    }
                  }}"""

DEFAULT_SG_RULES = {
    'myapp': [
        {
            'start_port': 22,
            'end_port': 22,
        },
    ],
    'test_app': [
        {
            'start_port': 31,
            'end_port': 31,
        },
    ],
}


@mock.patch('foremast.securitygroup.create_securitygroup.boto3')
@mock.patch('foremast.securitygroup.create_securitygroup.get_security_group_id')
@mock.patch('foremast.securitygroup.create_securitygroup.get_vpc_id')
@mock.patch('foremast.securitygroup.create_securitygroup.wait_for_task')
@mock.patch("foremast.securitygroup.create_securitygroup.get_properties")
@mock.patch("foremast.securitygroup.create_securitygroup.get_details")
def test_create_crossaccount_securitygroup(get_details, pipeline_config, wait_for_task, get_vpc_id,
                                           get_security_group_id, boto3):
    """Should create SG with cross account true"""
    pipeline_config.return_value = json.loads(SAMPLE_JSON)

    get_security_group_id.return_value = 'SGID'
    get_vpc_id.return_value = 'VPCID'

    x = SpinnakerSecurityGroup(app='edgeforrest', env='dev', region='us-east-1')
    assert x.create_security_group() is True

    no_cross_account_data = {'end_port': 8080, 'env': 'dev', 'protocol': 'tcp', 'start_port': 8080}
    no_cross_account_result = {'app': 'edgeforrest', 'end_port': 8080, 'cross_account_env': None, 'protocol': 'tcp', 'start_port': 8080, 'cross_account_vpc_id': None}
    no_cross_account = x.create_ingress_rule(app='edgeforrest', rule=no_cross_account_data)
    assert no_cross_account == no_cross_account_result

    cross_account_data = {'end_port': 8080, 'env': 'stage', 'protocol': 'tcp', 'start_port': 8080}
    cross_account_result = {'app': 'edgeforrest', 'end_port': 8080, 'cross_account_env': 'stage', 'protocol': 'tcp', 'start_port': 8080, 'cross_account_vpc_id': 'VPCID'}
    cross_account = x.create_ingress_rule(app='edgeforrest', rule=cross_account_data)
    assert cross_account == cross_account_result

    no_cross_account_simple = x.create_ingress_rule(app='edgeforrest', rule=8080)
    assert no_cross_account_simple == no_cross_account_result


@mock.patch('foremast.securitygroup.create_securitygroup.get_security_group_id')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
@mock.patch("foremast.securitygroup.create_securitygroup.get_details")
def test_missing_configuration(get_details, get_properties, get_sec_id):
    """Make missing Security Group configurations more apparent."""
    get_properties.return_value = {}

    security_group = SpinnakerSecurityGroup()

    with pytest.raises(ForemastConfigurationFileError):
        security_group.create_security_group()


@mock.patch('foremast.securitygroup.create_securitygroup.get_security_group_id')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
@mock.patch("foremast.securitygroup.create_securitygroup.get_details")
def test_misconfiguration(get_details, get_properties, get_sec_id):
    """Make bad Security Group definitions more apparent."""
    get_properties.return_value = {'security_group': {}}

    security_group = SpinnakerSecurityGroup()

    with pytest.raises(ForemastConfigurationFileError):
        security_group.create_security_group()


@mock.patch('foremast.securitygroup.create_securitygroup.boto3')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
@mock.patch('foremast.securitygroup.create_securitygroup.get_security_group_id')
@mock.patch("foremast.securitygroup.create_securitygroup.get_details")
def test_tags(get_details, get_security_group_id, get_properties, boto3):
    """Make bad Security Group definitions more apparent."""
    get_properties.return_value = {'security_group': {}}
    get_security_group_id.return_value = 'SGID'

    security_group = SpinnakerSecurityGroup()
    assert security_group.add_tags() is True


@mock.patch('foremast.securitygroup.create_securitygroup.get_details')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
def test_default_security_groups(mock_properties, mock_details):
    """Make sure default Security Groups are added to the ingress rules."""
    ingress = {
        'test_app': [
            {
                'start_port': 30,
                'end_port': 30,
            },
        ],
    }

    mock_properties.return_value = {
        'security_group': {
            'ingress': ingress,
            'description': '',
        },
    }

    test_sg = {
        'myapp': [
            {
                'start_port': '22',
                'end_port': '22',
                'protocol': 'tcp'
            },
        ]
    }
    with mock.patch.dict('foremast.securitygroup.create_securitygroup.DEFAULT_SECURITYGROUP_RULES', test_sg):
        sg = SpinnakerSecurityGroup()
        ingress = sg.update_default_rules()
        assert 'myapp' in ingress


@mock.patch('foremast.securitygroup.create_securitygroup.get_details')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
def test_securitygroup_references(mock_properties, mock_details):
    """Make sure default Security Groups are added to the ingress rules."""
    test_sg = {
        '$self': [
            {
                'start_port': '22',
                'end_port': '22',
                'protocol': 'tcp'
            },
        ]
    }

    sg = SpinnakerSecurityGroup(app='myapp')
    ingress = sg.resolve_self_references(test_sg)
    assert 'myapp' in ingress
    assert '22' == ingress['myapp'][0]['start_port']
    assert '22' == ingress['myapp'][0]['end_port']


@mock.patch.dict('foremast.securitygroup.create_securitygroup.DEFAULT_SECURITYGROUP_RULES', DEFAULT_SG_RULES)
@mock.patch('foremast.securitygroup.create_securitygroup.get_details')
@mock.patch('foremast.securitygroup.create_securitygroup.get_properties')
def test_merge_security_groups(mock_properties, mock_details):
    """Make sure default Security Groups are added to the ingress rules."""
    app_ingress = {
        'test_app': [
            {
                'start_port': 30,
                'end_port': 30,
            },
        ],
    }

    mock_properties.return_value = {
        'security_group': {
            'ingress': app_ingress,
            'description': '',
        },
    }

    sg = SpinnakerSecurityGroup()
    ingress = sg.update_default_rules()
    assert ingress['myapp'][0]['start_port'] == 22
    assert ingress['test_app'][0]['start_port'] == 31
    assert ingress['test_app'][1]['start_port'] == 30



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Cut Health Target up into pieces."""
import collections
import logging

LOG = logging.getLogger(__name__)


def splay_health(health_target):
    """Set Health Check path, port, and protocol.

    Args:
        health_target (str): The health target. ie ``HTTP:80``
    Returns:
        HealthCheck: A **collections.namedtuple** class with *path*, *port*,
        *proto*, and *target* attributes.
    """
    HealthCheck = collections.namedtuple('HealthCheck', ['path', 'port', 'proto', 'target'])

    proto, health_port_path = health_target.split(':')
    port, *health_path = health_port_path.split('/')

    if proto == 'TCP':
        path = ''
    elif not health_path:
        path = '/healthcheck'
    else:
        path = '/{0}'.format('/'.join(health_path))

    target = '{0}:{1}{2}'.format(proto, port, path)

    health = HealthCheck(path, port, proto, target)
    LOG.info(health)

    return health



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create manual Pipeline for Spinnaker."""
import json
import jinja2


from ..consts import TEMPLATES_PATH
from ..utils import get_pipeline_id, normalize_pipeline_name, get_jinja_environment
from ..utils.lookups import FileLookup
from .create_pipeline import SpinnakerPipeline
from .jinja_functions import get_jinja_functions, get_jinja_variables


class SpinnakerPipelineManual(SpinnakerPipeline):
    """Manual JSON configured Spinnaker Pipelines."""

    def create_pipeline(self):
        """Use JSON files to create Pipelines."""

        pipelines = self.settings['pipeline']['pipeline_files']
        self.log.info('Uploading manual Pipelines: %s', pipelines)

        for i, file_name in enumerate(pipelines):

            json_string = self.get_pipeline_file_contents(file_name)

            # If this is a .j2 file render is template, otherwise treat as normal json file
            is_jinja_template = False
            if file_name.endswith(".j2"):
                is_jinja_template = True
                pipeline_vars = self.get_pipeline_variables_dict(i)
                try:
                    json_string = self.get_rendered_json(json_string, pipeline_vars).strip()
                # Catch all exceptions (we don't know what errors the jinja2 user code may throw)
                # Add logging, then re-raise
                except Exception:
                    log_message = "Exception raised during Jinja 2 template rendering.  Check syntax in templates."
                    self.log.exception(log_message)
                    raise
                self.log.info("Jinja2 template successfully rendered")

            try:
                pipelines = json.loads(json_string)
                # result may be a single pipeline object, or list of pipeline objects
                # if a single pipeline object, transform to list with one pipeline object inside
                if not isinstance(pipelines, list):
                    self.log.debug("single pipeline object rendered, transforming to list")
                    pipelines = [pipelines]

            except json.JSONDecodeError:
                # Helpful debugging logging, then re-raise fatal exception
                if is_jinja_template:
                    self.log.error("JSONDecodeError parsing rendered Jinja template, verify it produces valid json")
                    self.log.debug(json_string)
                else:
                    self.log.error("JSONDecodeError parsing json, verify template is vaild json")
                raise

            # Create all pipelines
            for pipeline_dict in pipelines:
                pipeline_dict.setdefault('application', self.app_name)
                pipeline_dict.setdefault('name',
                                         normalize_pipeline_name(name=file_name.lstrip("templates://")))
                pipeline_dict.setdefault('id', get_pipeline_id(app=pipeline_dict['application'],
                                                               name=pipeline_dict['name']))
                self.post_pipeline(pipeline_dict)

        return True

    def get_pipeline_file_contents(self, file_name):
        """Returns a string containing the file constants of the file passed in
        Supports local files, files in git and shared templates in the TEMPLATES_PATH

        Args:
            file_name (str): pipeline file name

        Returns:
            str: Contents of pipeline file."""
        # Check if this file is a shared template in the TEMPLATES_PATH
        if file_name.startswith("templates://"):
            if TEMPLATES_PATH is None:
                raise Exception("Cannot use templates:// schema without TEMPLATES_PATH")
            # Using lstrip or strip can sometimes remove additional chars
            # We know the string starts with "templates://" already, so remove the exact
            # number of chars to be safe
            schema_length = len("templates://")
            file_name = file_name[schema_length::]
            self.log.debug("Updated pipeline template file path '%s'", file_name)
            pipeline_templates_path = TEMPLATES_PATH.rstrip("/") + "/pipeline"
            lookup = FileLookup(git_short=None, runway_dir=pipeline_templates_path)
        else:
            # Consider it a local repo file, check local or git:
            lookup = FileLookup(git_short=self.generated.gitlab()['main'], runway_dir=self.runway_dir)

        return lookup.get(filename=file_name)

    def get_rendered_json(self, json_string, pipeline_vars=None):
        """Takes a string of a manual template and renders it as a Jinja2 template, returning the result

        Args:
            json_string (str): pipeline in jinja/json format
            pipeline_vars (dict): key/value pairs of variables the pipline expects

        Returns:
            str: pipeline json after Jinja is rendered"""

        jinja_env = get_jinja_environment()

        try:
            jinja_template = jinja_env.from_string(json_string)
            # Get any pipeline args defined in pipeline.json, default to empty dict if none defined
            pipeline_args = dict()
        except jinja2.TemplateNotFound:
            # Log paths searched for debugging, then re-raise
            message = 'Jinja2 TemplateNotFound exception in paths {paths}'.format(paths=jinja_env.loader.searchpath)
            self.log.error(message)
            raise

        # Expose permitted functions and variables to the template
        pipeline_args.update(get_jinja_functions())
        pipeline_args.update(get_jinja_variables(self))

        # If any args set in the pipeline file add them to the pipeline_args.variables
        if pipeline_vars is not None:
            pipeline_args["template_variables"] = pipeline_vars

        # Render the template
        return jinja_template.render(pipeline_args)

    def get_pipeline_variables_dict(self, index):
        """Safely gets the user defined variables from the pipeline.json file

        Args:
            index (int): index of pipeline defined in pipeline.json

        Returns:
            dict: key/value pair of variables defined in pipeline.json"""

        # Safely get variables out of pipeline.json if they are set
        pipe_settings = self.settings["pipeline"]
        if "template_variables" in pipe_settings \
           and isinstance(pipe_settings["template_variables"], list) \
           and len(pipe_settings["template_variables"]) > index:
            return pipe_settings["template_variables"][index]

        # Default value is None
        return None



"""CloudWatch Event functions."""
import logging

import boto3

from ..utils.awslambda import get_lambda_alias_arn

LOG = logging.getLogger(__name__)


def get_cloudwatch_event_rule(app_name, account, region):
    """Get CloudWatch Event rule names."""
    session = boto3.Session(profile_name=account, region_name=region)
    cloudwatch_client = session.client('events')

    lambda_alias_arn = get_lambda_alias_arn(app=app_name, account=account, region=region)
    rule_names = cloudwatch_client.list_rule_names_by_target(TargetArn=lambda_alias_arn)

    if rule_names['RuleNames']:
        all_rules = rule_names['RuleNames']
    else:
        LOG.debug("No event rules found")
        all_rules = []
    return all_rules



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Write output files for configurations."""
import json
import logging
from pprint import pformat

import foremastutils

from ..consts import APP_FORMATS
from ..utils import DeepChainMap, get_template

LOG = logging.getLogger(__name__)


def convert_ini(config_dict):
    """Convert _config_dict_ into a list of INI formatted strings.

    Args:
        config_dict (dict): Configuration dictionary to be flattened.

    Returns:
        (list) Lines to be written to a file in the format of KEY1_KEY2=value.
    """
    config_lines = []

    for env, configs in sorted(config_dict.items()):
        for resource, app_properties in sorted(configs.items()):
            try:
                for app_property, value in sorted(app_properties.items()):
                    variable = '{env}_{resource}_{app_property}'.format(
                        env=env, resource=resource, app_property=app_property).upper()

                    if isinstance(value, (dict, DeepChainMap)):
                        safe_value = "'{0}'".format(json.dumps(dict(value)))
                    else:
                        safe_value = json.dumps(value)

                    line = "{variable}={value}".format(variable=variable, value=safe_value)

                    LOG.debug('INI line: %s', line)
                    config_lines.append(line)
            except AttributeError:
                resource = resource.upper()
                app_properties = "'{}'".format(json.dumps(app_properties))
                line = '{0}={1}'.format(resource, app_properties)

                LOG.debug('INI line: %s', line)
                config_lines.append(line)
    return config_lines


def write_variables(app_configs=None, out_file='', git_short=''):
    """Append _application.json_ configs to _out_file_, .exports, and .json.

    Variables are written in INI style, e.g. UPPER_CASE=value. The .exports file
    contains 'export' prepended to each line for easy sourcing. The .json file
    is a minified representation of the combined configurations.

    Args:
        app_configs (dict): Environment configurations from _application.json_
            files, e.g. {'dev': {'elb': {'subnet_purpose': 'internal'}}}.
        out_file (str): Name of INI file to append to.
        git_short (str): Short name of Git repository, e.g. forrest/core.

    Returns:
        dict: Configuration equivalent to the JSON output.
    """
    generated = foremastutils.Generator(*foremastutils.Parser(git_short).parse_url(), formats=APP_FORMATS)

    json_configs = {}
    for env, configs in app_configs.items():
        if env != 'pipeline':
            instance_profile = generated.iam()['profile']
            rendered_configs = json.loads(
                get_template(
                    'configs/configs.json.j2',
                    env=env,
                    app=generated.app_name(),
                    profile=instance_profile,
                    formats=generated))
            json_configs[env] = dict(DeepChainMap(configs, rendered_configs))
            region_list = configs.get('regions', rendered_configs['regions'])
            json_configs[env]['regions'] = region_list  # removes regions defined in templates but not configs.
            for region in region_list:
                region_config = json_configs[env][region]
                json_configs[env][region] = dict(DeepChainMap(region_config, rendered_configs))
        else:
            default_pipeline_json = json.loads(get_template('configs/pipeline.json.j2', formats=generated))
            json_configs['pipeline'] = dict(DeepChainMap(configs, default_pipeline_json))

    LOG.debug('Compiled configs:\n%s', pformat(json_configs))

    config_lines = convert_ini(json_configs)

    with open(out_file, 'at') as jenkins_vars:
        LOG.info('Appending variables to %s.', out_file)
        jenkins_vars.write('\n'.join(config_lines))

    with open(out_file + '.exports', 'wt') as export_vars:
        LOG.info('Writing sourceable variables to %s.', export_vars.name)
        export_vars.write('\n'.join('export {0}'.format(line) for line in config_lines))

    with open(out_file + '.json', 'wt') as json_handle:
        LOG.info('Writing JSON to %s.', json_handle.name)
        LOG.debug('Total JSON dict:\n%s', json_configs)
        json.dump(json_configs, json_handle)

    return json_configs



"""dynamodb table stream functions."""
import logging

import boto3

from ..exceptions import DynamoDBStreamNotFound, DynamoDBTableNotFound

LOG = logging.getLogger(__name__)


def check_arn_type(arn_string):
    """Check ARN

    Args:
        arn_string (str): ARN String to check and validate

    Returns:
        str: ARN for requested table name

    """

    if arn_string.startswith('arn:aws:dynamodb:'):
        _prefix, table, *stream = arn_string.split('/')

        LOG.debug("ARN Split: %s %s %s", _prefix, table, stream)
        if stream:
            arn_type = "dynamodb-stream"
        elif table:
            arn_type = "dynamodb-table"
        else:
            raise Exception('ARN Type could not be determined from {0}. Check provided ARN String.'.format(arn_string))

    return arn_type


def lookup_latest_dynamodb_stream(account, region, arn_string=None, table_name=None):
    """Lookup dynamodb stream ARN by DynamoDB table arn or raw table name

    Args:
        account (str): Environment, e.g. dev
        region (str): Region name, e.g. us-east-1
        arn_string (str): DynamoDB ARN String to look for stream
        table_name (str): DynamoDB table name to look for stream

    Returns:
        str: ARN for latest DynamoDB stream ARN

    """
    session = boto3.Session(profile_name=account, region_name=region)
    dynamodb_client = session.client('dynamodb')
    dynamodb_streams_client = session.client('dynamodbstreams')

    if arn_string:
        _prefix, table_name, *stream = arn_string.split('/')
        LOG.info('DynamoDB Stream ARN provided. Looking up Table Name from: {0}'.format(arn_string))

    LOG.info('Checking DynamoDB Table Response of table {0}'.format(table_name))
    table_response = dynamodb_client.describe_table(TableName=table_name)

    if 'Table' not in table_response:
        LOG.critical("No DynamoDB table named %s was found.", table_name)
        raise DynamoDBTableNotFound('No DynamoDB table named {0} was found'.format(table_name))

    streams_response = dynamodb_streams_client.list_streams(TableName=table_name)
    streams = streams_response['Streams']

    # Return latest stream if exact stream not provided
    try:
        latest_stream_arn = streams[0]['StreamArn']
    except IndexError:
        LOG.critical("No DynamoDB stream found for table %s.", table_name)
        raise DynamoDBStreamNotFound('No DynamoDB stream found for table named {0}'.format(table_name))
    return latest_stream_arn


def get_dynamodb_stream_arn(arn_string, account, region):
    """Get DynamoDB stream ARN from a DynamoDB table.

    Args:
        arn_string (str): Name of the table stream to lookup a stream
        account (str): Environment, e.g. dev
        region (str): Region name, e.g. us-east-1

    Returns:
        str: ARN for requested table name

    """
    arn_type = check_arn_type(arn_string)

    if arn_type == 'dynamodb-stream':
        return arn_string
    elif arn_type == 'dynamodb-table':
        return lookup_latest_dynamodb_stream(account, region, arn_string=arn_string)
    return lookup_latest_dynamodb_stream(account, region, table_name=arn_string)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""ChainMap modification to handle nested dict objects."""
import collections


# pylint: disable=too-many-ancestors
class DeepChainMap(collections.ChainMap):
    """Deep lookups for collections.ChainMap objects.

    When there are nested dicts, the first found, second level dict is returned
    instead of overlaying alternative second level dicts.

        >>> first = {'key1': {'key1_1': 'first_one'}}
        >>> second = {'key1': {'key1_1': 'second_one', 'key1_2': 'second_two'}}
        >>> collections.ChainMap(first, second)['key1']
        {'key1_1': 'first_one'}
        >>> collections.ChainMap(second, first)['key1']
        {'key1_1': 'second_one', 'key1_2': 'second_two'}

        # Deep lookup will flatten every level.

        >>> DeepChainMap(first, second)['key1']
        {'key1_1': 'first_one', 'key1_2': 'second_two'}
        >>> DeepChainMap(second, first)['key1']
        {'key1_1': 'second_one', 'key1_2': 'second_two'}
    """

    def __getitem__(self, key):
        """Recursively retrieve value for _key_ in dict.

        Args:
            key (str): dict key to get all items for
        """
        for mapping in self.maps:
            try:
                value = mapping[key]
                map_value = value
                if isinstance(value, dict):
                    map_value = dict(DeepChainMap(*list(mapping.get(key, {}) for mapping in self.maps)))
                return map_value
            except KeyError:
                pass
        return self.__missing__(key)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""CLI entry point for aws events creation.

Help: ``python -m src.foremast.awslambdaevent -h``
"""
import argparse
import logging

from ..args import add_app, add_debug, add_env, add_properties, add_region
from ..consts import LOGGING_FORMAT
from .awslambda import LambdaFunction
from .awslambdaevent import LambdaEvent


def main():
    """Create Lambda events."""
    logging.basicConfig(format=LOGGING_FORMAT)
    log = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(description=main.__doc__)
    add_debug(parser)
    add_app(parser)
    add_env(parser)
    add_properties(parser)
    add_region(parser)
    args = parser.parse_args()

    logging.getLogger(__package__.split('.')[0]).setLevel(args.debug)

    log.debug('Parsed arguments: %s', args)

    lambda_function = LambdaFunction(app=args.app, env=args.env, region=args.region, prop_path=args.properties)

    lambda_function.create_lambda_function()

    lambda_event = LambdaEvent(app=args.app, env=args.env, region=args.region, prop_path=args.properties)
    lambda_event.create_lambda_events()


if __name__ == "__main__":
    main()



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Get Application properties that have been generated by `create-configs`."""
import json
import logging

LOG = logging.getLogger(__name__)


def get_properties(properties_file='raw.properties.json', env=None, region=None):
    """Get contents of _properties_file_ for the _env_.

    Args:
        properties_file (str): File name of `create-configs` JSON output.
        env (str): Environment to read optionally.
        region (str): Region to get specific configs for.

    Returns:
        dict: JSON loaded Application properties for _env_.
        None: Given _env_ was not found in `create-configs` JSON output.

    """
    with open(properties_file, 'rt') as file_handle:
        properties = json.load(file_handle)

    env_properties = properties.get(env, properties)
    contents = env_properties.get(region, env_properties)
    LOG.debug('Found properties for %s:\n%s', env, contents)
    return contents



"""Verifies that instance_links are being retrieved properly from LINKS. Verifies that app_data.json.j2
contains the instance link information"""
from unittest import mock
from foremast import configs

TEST_CONFIG =  { 
    "asg": {
        "subnet_purpose": "internal",
        "min_inst": 3,
        "max_inst": 3
    },
    "regions": {
        "us-east-1": {},
        "us-west-2": {
            "asg": {
                "min_inst": 1,
                "max_inst": 1
            }
        }
    }
}

def test_appy_region_configs():
    """Tests that `configs.apply_region_configs` applies
    region specific overrides correctly""" 
    desired_config = { 
        "asg": {
            "subnet_purpose": "internal",
            "min_inst": 3,
            "max_inst": 3
        },
        "regions": {
            "us-east-1": {},
            "us-west-2": {
                "asg": {
                    "min_inst": 1,
                    "max_inst": 1
                }
            }
        },
        "us-east-1": { 
            "asg": {
                "subnet_purpose": "internal",
                "min_inst": 3,
                "max_inst": 3
            },
            "regions": {
                "us-east-1": {},
                "us-west-2": {
                    "asg": {
                        "min_inst": 1,
                        "max_inst": 1
                    }
                }
            }
        },
        "us-west-2": { 
            "asg": {
                "subnet_purpose": "internal",
                "min_inst": 1,
                "max_inst": 1
            },
            "regions": {
                "us-east-1": {},
                "us-west-2": {
                    "asg": {
                        "min_inst": 1,
                        "max_inst": 1
                    }
                }
            }
        }
    }
    rendered_config = configs.apply_region_configs(TEST_CONFIG)
    assert rendered_config == desired_config

def test_legacy_region_format():
    """Validates that the only regions list format continues to work."""
    desired_config = {
        "asg": {
            "subnet_purpose": "internal",
            "min_inst": 3,
            "max_inst": 3
        },
        "regions": ["us-east-1", "us-west-2"],
        "us-east-1": {
            "asg": {
                "subnet_purpose": "internal",
                "min_inst": 3,
                "max_inst": 3
            },
            "regions": ["us-east-1", "us-west-2"],
        },
        "us-west-2": {
            "asg": {
                "subnet_purpose": "internal",
                "min_inst": 3,
                "max_inst": 3
            },
            "regions": ["us-east-1", "us-west-2"],
        },
    }
    legacy_config = TEST_CONFIG.copy()
    legacy_config["regions"] = ['us-east-1', 'us-west-2']
    rendered_config = configs.apply_region_configs(legacy_config)
    assert rendered_config == desired_config



    



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from unittest.mock import MagicMock, patch

from foremast import dns


@patch('foremast.dns.create_dns.update_dns_zone_record')
@patch('foremast.dns.create_dns.get_dns_zone_ids')
@patch('foremast.dns.create_dns.find_elb')
@patch('foremast.dns.create_dns.get_properties')
@patch('foremast.dns.create_dns.DOMAIN', 'example.com')
@patch('foremast.dns.create_dns.get_details')
def test_dns_creation(mock_get_details, mock_properties, mock_find_elb, mock_dns_zones, mock_update_dns):
    # mocked data
    hosted_zones = [500, 501]
    dns_elb = {'elb': 'myapp.dev1.example.com'}

    # mock results
    mock_get_details.return_value.app_name.return_value = 'myapp'
    mock_get_details.return_value.dns.return_value = dns_elb
    mock_properties.return_value = {'dns': {'ttl': 60}}
    mock_find_elb.return_value = 'myapp-internal.domain.external.com'
    mock_dns_zones.return_value = hosted_zones
    mock_update_dns.return_value = True

    d = dns.SpinnakerDns(app='myapp', region='region1', env='dev1', elb_subnet='noexist')
    assert d.domain == 'example.com'
    assert d.region == 'region1'
    assert d.elb_subnet == 'noexist'
    assert d.app_name == 'myapp'

    assert 'myapp.dev1.example.com' == d.create_elb_dns()

    sent_update_data = {
        'dns_name': dns_elb['elb'],
        'dns_name_aws': mock_find_elb.return_value,
        'dns_ttl': mock_properties.return_value['dns']['ttl']
    }

    mock_update_dns.assert_called_with('dev1', 501, **sent_update_data)
    assert mock_update_dns.call_count == 2



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Render Jinja2 template."""
import logging
import os
import pathlib

import jinja2

from ..consts import TEMPLATES_PATH
from ..exceptions import ForemastTemplateNotFound

LOG = logging.getLogger(__name__)

HERE = pathlib.Path(__file__).parent.absolute()
LOCAL_TEMPLATES = HERE.joinpath('../templates/').resolve()


def get_jinja_environment():
    """Gets the Foremast Jinja environment used for rendering templates
    Returns:
        jinja2.Environment
    """
    jinja_template_paths_obj = []

    if TEMPLATES_PATH:
        external_templates = pathlib.Path(TEMPLATES_PATH).expanduser().resolve()
        assert os.path.isdir(external_templates), 'External template path "{0}" not found'.format(external_templates)
        jinja_template_paths_obj.append(external_templates)

    jinja_template_paths_obj.append(LOCAL_TEMPLATES)
    jinja_template_paths = [str(path) for path in jinja_template_paths_obj]

    jinjaenv = jinja2.Environment(loader=jinja2.FileSystemLoader(jinja_template_paths))
    return jinjaenv


def get_template_object(template_file=''):
    """Retrieve template.

    Args:
        template_file (str): Name of template file.

    Returns:
        jinja2.Template: Template ready to render.

    Raises:
        AssertionError: Configured path for templates does not exist.
        :obj:`foremast.exceptions.ForemastTemplateNotFound`: Requested template
            is not available.

    """
    jinjaenv = get_jinja_environment()

    try:
        template = jinjaenv.get_template(template_file)
    except jinja2.TemplateNotFound:
        message = 'Unable to find template "{template_file}" in paths {paths}'.format(
            template_file=template_file, paths=jinjaenv.loader.searchpath)
        LOG.error(message)
        raise ForemastTemplateNotFound(message)

    return template


def get_template(template_file='', **kwargs):
    """Get the Jinja2 template and renders with dict _kwargs_.

    Args:
        template_file (str): name of the template file
        kwargs: Keywords to use for rendering the Jinja2 template.

    Returns:
        String of rendered JSON template.

    """
    template = get_template_object(template_file)

    LOG.info('Rendering template %s', template.filename)
    for key, value in kwargs.items():
        LOG.debug('%s => %s', key, value)

    rendered_json = template.render(**kwargs)
    LOG.debug('Rendered JSON:\n%s', rendered_json)

    return rendered_json



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy lambda sns events"""

import logging

import boto3

from ....utils.get_sns_subscriptions import get_sns_subscriptions

LOG = logging.getLogger(__name__)


def destroy_sns_event(app_name, env, region):
    """ Destroy all Lambda SNS subscriptions.

    Args:
        app_name (str): name of the lambda function
        env (str): Environment/Account for lambda function
        region (str): AWS region of the lambda function

    Returns:
        boolean: True if subscription destroyed successfully
    """
    session = boto3.Session(profile_name=env, region_name=region)
    sns_client = session.client('sns')

    lambda_subscriptions = get_sns_subscriptions(app_name=app_name, env=env, region=region)

    for subscription_arn in lambda_subscriptions:
        sns_client.unsubscribe(SubscriptionArn=subscription_arn)

    LOG.debug("Lambda SNS event deleted")
    return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""S3 web application infrastructure."""
import json
import logging

import boto3
from botocore.client import ClientError

from ..exceptions import S3SharedBucketNotFound
from ..utils import generate_s3_tags, get_details, get_dns_zone_ids, get_properties, update_dns_zone_record

LOG = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class S3Apps:
    """Configure infrastructure and policies for S3 web applications."""

    def __init__(self, app, env, region, prop_path, primary_region='us-east-1'):
        """S3 application object. Setups Bucket and policies for S3 applications.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
            primary_region (str): The primary region for the application.
        """
        self.app_name = app
        self.env = env
        self.region = region
        boto_sess = boto3.session.Session(profile_name=env)
        self.s3client = boto_sess.client('s3')
        self.generated = get_details(app=app, env=env, region=self.region)
        self.properties = get_properties(prop_path, env=self.env, region=self.region)
        self.s3props = self.properties['s3']
        self.group = self.generated.project

        include_region = True
        if self.region == primary_region:
            include_region = False
        if self.s3props.get('shared_bucket_master'):
            self.bucket = self.generated.shared_s3_app_bucket(include_region=include_region)
        elif self.s3props.get('shared_bucket_target'):
            shared_app = self.s3props['shared_bucket_target']
            newgenerated = get_details(app=shared_app, env=env, region=self.region)
            self.bucket = newgenerated.shared_s3_app_bucket(include_region=include_region)
        elif self.s3props.get('bucket_name'):
            self.bucket = self.s3props['bucket_name']
        else:
            self.bucket = self.generated.s3_app_bucket(include_region=include_region)

    def create_bucket(self):
        """Create or update bucket based on app name."""
        bucket_exists = self._bucket_exists()
        if self.s3props.get('shared_bucket_target'):
            if bucket_exists:
                LOG.info('App uses shared bucket - %s ', self.bucket)
            else:
                LOG.error("Shared bucket %s does not exist", self.bucket)
                raise S3SharedBucketNotFound
        else:
            if self.region == 'us-east-1':
                _response = self.s3client.create_bucket(ACL=self.s3props['bucket_acl'], Bucket=self.bucket)
            else:
                if not bucket_exists:
                    _response = self.s3client.create_bucket(ACL=self.s3props['bucket_acl'], Bucket=self.bucket,
                                                            CreateBucketConfiguration={
                                                                'LocationConstraint': self.region})
                else:
                    _response = "bucket already exists, skipping create for non-standard region buckets."
            LOG.debug('Response creating bucket: %s', _response)
            LOG.info('%s - S3 Bucket Upserted', self.bucket)
            self._put_bucket_policy()
            self._put_bucket_website()
            self._put_bucket_logging()
            self._put_bucket_lifecycle()
            self._put_bucket_versioning()
            self._put_bucket_encryption()
            self._put_bucket_notification()
            self._put_bucket_tagging()

    def _bucket_exists(self):
        """Check if the bucket exists."""
        try:
            self.s3client.get_bucket_location(Bucket=self.bucket)
            return True
        except ClientError as error:
            LOG.error(error)
            return False

    def _put_bucket_policy(self):
        """Attach a bucket policy to app bucket."""
        if self.s3props['bucket_policy']:
            policy_str = json.dumps(self.s3props['bucket_policy'])
            _response = self.s3client.put_bucket_policy(Bucket=self.bucket, Policy=policy_str)
        else:
            _response = self.s3client.delete_bucket_policy(Bucket=self.bucket)
        LOG.debug('Response adding bucket policy: %s', _response)
        LOG.info('S3 Bucket Policy Attached')

    def _put_bucket_website(self):
        """Configure static website on S3 bucket."""
        if self.s3props['website']['enabled']:
            website_config = {
                'ErrorDocument': {
                    'Key': self.s3props['website']['error_document']
                },
                'IndexDocument': {
                    'Suffix': self.s3props['website']['index_suffix']
                }
            }
            _response = self.s3client.put_bucket_website(Bucket=self.bucket, WebsiteConfiguration=website_config)
            self._put_bucket_cors()
            self._set_bucket_dns()
        else:
            _response = self.s3client.delete_bucket_website(Bucket=self.bucket)
            self._put_bucket_cors()
        LOG.debug('Response setting up S3 website: %s', _response)
        LOG.info('S3 website settings updated')

    def _set_bucket_dns(self):
        """Create CNAME for S3 endpoint."""
        # Different regions have different s3 endpoint formats
        dotformat_regions = ["eu-west-2", "eu-central-1", "ap-northeast-2", "ap-south-1", "ca-central-1", "us-east-2"]
        if self.region in dotformat_regions:
            s3_endpoint = "{0}.s3-website.{1}.amazonaws.com".format(self.bucket, self.region)
        else:
            s3_endpoint = "{0}.s3-website-{1}.amazonaws.com".format(self.bucket, self.region)

        zone_ids = get_dns_zone_ids(env=self.env, facing="public")
        dns_kwargs = {
            'dns_name': self.bucket,
            'dns_name_aws': s3_endpoint,
            'dns_ttl': self.properties['dns']['ttl']
        }

        for zone_id in zone_ids:
            LOG.debug('zone_id: %s', zone_id)
            update_dns_zone_record(self.env, zone_id, **dns_kwargs)
        LOG.info("Created DNS %s for Bucket", self.bucket)

    def _put_bucket_cors(self):
        """Adds bucket cors configuration."""
        if self.s3props['cors']['enabled'] and self.s3props['website']['enabled']:
            cors_config = {}
            cors_rules = []
            for each_rule in self.s3props['cors']['cors_rules']:
                cors_rules.append({
                    'AllowedHeaders': each_rule['cors_headers'],
                    'AllowedMethods': each_rule['cors_methods'],
                    'AllowedOrigins': each_rule['cors_origins'],
                    'ExposeHeaders': each_rule['cors_expose_headers'],
                    'MaxAgeSeconds': each_rule['cors_max_age']
                })
            cors_config = {
                'CORSRules': cors_rules
            }
            LOG.debug(cors_config)
            _response = self.s3client.put_bucket_cors(Bucket=self.bucket, CORSConfiguration=cors_config)
        else:
            _response = self.s3client.delete_bucket_cors(Bucket=self.bucket)
        LOG.debug('Response setting up S3 CORS: %s', _response)
        LOG.info('S3 CORS configuration updated')

    def _put_bucket_encryption(self):
        """Adds bucket encryption configuration."""
        if self.s3props['encryption']['enabled']:
            encryption_config = {'Rules': [{}]}
            encryption_config = {
                'Rules': self.s3props['encryption']['encryption_rules']
            }
            LOG.debug(encryption_config)
            _response = self.s3client.put_bucket_encryption(Bucket=self.bucket,
                                                            ServerSideEncryptionConfiguration=encryption_config)
        else:
            _response = self.s3client.delete_bucket_encryption(Bucket=self.bucket)
        LOG.debug('Response setting up S3 encryption: %s', _response)
        LOG.info('S3 encryption configuration updated')

    def _put_bucket_lifecycle(self):
        """Adds bucket lifecycle configuration."""
        status = 'deleted'
        if self.s3props['lifecycle']['enabled']:
            lifecycle_config = {
                'Rules': self.s3props['lifecycle']['lifecycle_rules']
            }
            LOG.debug('Lifecycle Config: %s', lifecycle_config)
            _response = self.s3client.put_bucket_lifecycle_configuration(Bucket=self.bucket,
                                                                         LifecycleConfiguration=lifecycle_config)
            status = 'applied'
        else:
            _response = self.s3client.delete_bucket_lifecycle(Bucket=self.bucket)
        LOG.debug('Response setting up S3 lifecycle: %s', _response)
        LOG.info('S3 lifecycle configuration %s', status)

    def _put_bucket_logging(self):
        """Adds bucket logging policy to bucket for s3 access requests"""
        logging_config = {}
        if self.s3props['logging']['enabled']:
            logging_config = {
                'LoggingEnabled': {
                    'TargetBucket': self.s3props['logging']['logging_bucket'],
                    'TargetGrants': self.s3props['logging']['logging_grants'],
                    'TargetPrefix': self.s3props['logging']['logging_bucket_prefix']
                }
            }
        _response = self.s3client.put_bucket_logging(Bucket=self.bucket, BucketLoggingStatus=logging_config)
        LOG.debug('Response setting up S3 logging: %s', _response)
        LOG.info('S3 logging configuration updated')

    def _put_bucket_notification(self):
        """Adds bucket notification configuration."""
        notification_config = {}

        if self.s3props['notification']['enabled']:
            if 'topic_configurations' in self.s3props['notification'] and \
                    self.s3props['notification']['topic_configurations'] != [{}]:
                notification_config['TopicConfigurations'] = self.s3props['notification']['topic_configurations']
            if 'queue_configurations' in self.s3props['notification'] and \
                    self.s3props['notification']['queue_configurations'] != [{}]:
                notification_config['QueueConfigurations'] = self.s3props['notification']['queue_configurations']
            if 'lambda_configurations' in self.s3props['notification'] and \
                    self.s3props['notification']['lambda_configurations'] != [{}]:
                notification_config['LambdaFunctionConfigurations'] = \
                    self.s3props['notification']['lambda_configurations']
        LOG.debug('Notification Config: %s', notification_config)
        _response = self.s3client.put_bucket_notification_configuration(Bucket=self.bucket,
                                                                        NotificationConfiguration=notification_config)
        LOG.debug('Response setting up S3 notification: %s', _response)
        LOG.info('S3 notification configuration updated')

    def _put_bucket_tagging(self):
        """Add bucket tags to bucket."""
        app_block_tags = self.properties['app']['custom_tags']
        s3_block_tags = self.s3props['tagging']['tags']
        default_tags = {'app_group': self.group, 'app_name': self.app_name}

        s3_tags = {**default_tags, **s3_block_tags, **app_block_tags}
        tag_set = generate_s3_tags.generated_tag_data(s3_tags)

        tagging_config = {'TagSet': tag_set}

        self.s3client.put_bucket_tagging(Bucket=self.bucket, Tagging=tagging_config)
        LOG.info("Adding tagging %s for Bucket", tag_set)

    def _put_bucket_versioning(self):
        """Adds bucket versioning policy to bucket"""
        status = 'Suspended'
        if self.s3props['versioning']['enabled']:
            status = 'Enabled'

        versioning_config = {
            'MFADelete': self.s3props['versioning']['mfa_delete'],
            'Status': status
        }

        _response = self.s3client.put_bucket_versioning(Bucket=self.bucket, VersioningConfiguration=versioning_config)
        LOG.debug('Response setting up S3 versioning: %s', _response)
        LOG.info('S3 versioning configuration updated')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create Pipelines for Spinnaker."""
import collections
import json
from pprint import pformat

from ..consts import DEFAULT_RUN_AS_USER
from ..utils import get_subnets, get_template
from .clean_pipelines import clean_pipelines
from .construct_pipeline_block_lambda import construct_pipeline_block_lambda
from .create_pipeline import SpinnakerPipeline
from .renumerate_stages import renumerate_stages


class SpinnakerPipelineLambda(SpinnakerPipeline):
    """Manipulate Spinnaker Pipelines.

    Args:
        app (str): Application name.
        trigger_job (str): Jenkins trigger job.
        base (str): Base image name (i.e: fedora).
        prop_path (str): Path to the raw.properties.json.
    """

    def render_wrapper(self, region='us-east-1'):
        """Generate the base Pipeline wrapper.

        This renders the non-repeatable stages in a pipeline, like jenkins, baking, tagging and notifications.

        Args:
            region (str): AWS Region.

        Returns:
            dict: Rendered Pipeline wrapper.
        """
        base = self.settings['pipeline']['base']

        if self.base:
            base = self.base

        email = self.settings['pipeline']['notifications']['email']
        slack = self.settings['pipeline']['notifications']['slack']
        deploy_type = self.settings['pipeline']['type']
        pipeline_id = self.compare_with_existing(region=region)

        data = {
            'app': {
                'appname': self.app_name,
                'group_name': self.group_name,
                'repo_name': self.repo_name,
                'base': base,
                'deploy_type': deploy_type,
                'environment': 'packaging',
                'region': region,
                'triggerjob': self.trigger_job,
                'run_as_user': DEFAULT_RUN_AS_USER,
                'email': email,
                'slack': slack,
                'pipeline': self.settings['pipeline']
            },
            'id': pipeline_id
        }

        self.log.debug('Wrapper app data:\n%s', pformat(data))

        wrapper = get_template(template_file='pipeline/pipeline_wrapper.json.j2', data=data, formats=self.generated)

        return json.loads(wrapper)

    def create_pipeline(self):
        """Main wrapper for pipeline creation.
        1. Runs clean_pipelines to clean up existing ones
        2. determines which environments the pipeline needs
        3. gets all subnets for template rendering
        4. Renders all of the pipeline blocks as defined in configs
        5. Runs post_pipeline to create pipeline
        """
        clean_pipelines(app=self.app_name, settings=self.settings)

        pipeline_envs = self.environments
        self.log.debug('Envs from pipeline.json: %s', pipeline_envs)

        regions_envs = collections.defaultdict(list)
        for env in pipeline_envs:
            for region in self.settings[env]['regions']:
                regions_envs[region].append(env)
        self.log.info('Environments and Regions for Pipelines:\n%s', json.dumps(regions_envs, indent=4))

        subnets = get_subnets()

        pipelines = {}
        for region, envs in regions_envs.items():
            # TODO: Overrides for an environment no longer makes sense. Need to
            # provide override for entire Region possibly.
            pipelines[region] = self.render_wrapper(region=region)

            previous_env = None
            for env in envs:
                try:
                    region_subnets = {region: subnets[env][region]}
                except KeyError:
                    self.log.info('%s is not available for %s.', env, region)
                    continue

                block = construct_pipeline_block_lambda(
                    env=env,
                    generated=self.generated,
                    previous_env=previous_env,
                    region=region,
                    region_subnets=region_subnets,
                    settings=self.settings[env][region],
                    pipeline_data=self.settings['pipeline'])
                pipelines[region]['stages'].extend(json.loads(block))

                previous_env = env

        self.log.debug('Assembled Pipelines:\n%s', pformat(pipelines))

        for region, pipeline in pipelines.items():
            renumerate_stages(pipeline)

            self.post_pipeline(pipeline)

        return True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Application related utilities."""
import logging

import foremastutils

from ..consts import APP_FORMATS
from ..exceptions import SpinnakerAppNotFound
from ..utils.gate import gate_request

LOG = logging.getLogger(__name__)


def get_all_apps():
    """Get a list of all applications in Spinnaker.

    Returns:
        requests.models.Response: Response from Gate containing list of all apps.

    """
    LOG.info('Retreiving list of all Spinnaker applications')
    uri = '/applications'
    response = gate_request(uri=uri)

    assert response.ok, 'Could not retrieve application list'

    pipelines = response.json()
    LOG.debug('All Applications:\n%s', pipelines)

    return pipelines


def get_details(app='groupproject', env='dev', region='us-east-1'):
    """Extract details for Application.

    Args:
        app (str): Application Name
        env (str): Environment/account to get details from

    Returns:
        collections.namedtuple with _group_, _policy_, _profile_, _role_,
            _user_.

    """
    uri = '/applications/{app}'.format(app=app)
    request = gate_request(uri=uri)

    if not request.ok:
        raise SpinnakerAppNotFound('"{0}" not found.'.format(app))

    app_details = request.json()

    LOG.debug('App details: %s', app_details)
    group = app_details['attributes'].get('repoProjectKey')
    project = app_details['attributes'].get('repoSlug')
    generated = foremastutils.Generator(group, project, env=env, region=region, formats=APP_FORMATS)

    LOG.debug('Application details: %s', generated)
    return generated



"""Test IAM Policy templates are valid JSON."""
import json

import jinja2
import pytest

from foremast.iam.construct_policy import render_policy_template
from foremast.utils.templates import LOCAL_TEMPLATES


def iam_templates():
    """Generate list of IAM templates."""
    jinjaenv = jinja2.Environment(loader=jinja2.FileSystemLoader([str(LOCAL_TEMPLATES)]))

    iam_template_names = jinjaenv.list_templates(filter_func=lambda x: all([
        x.startswith('infrastructure/iam/'),
        'trust' not in x,
        'wrapper' not in x, ]))

    for iam_template_name in iam_template_names:
        yield iam_template_name


@pytest.mark.parametrize(argnames='template_name', argvalues=iam_templates())
def test_all_iam_templates(template_name):
    """Verify all IAM templates render as proper JSON."""
    *_, service_json = template_name.split('/')
    service, *_ = service_json.split('.')

    items = ['resource1', 'resource2']

    if service == 'rds-db':
        items = {
            'resource1': 'user1',
            'resource2': 'user2',
        }

    try:
        rendered = render_policy_template(
            account_number='',
            app='coreforrest',
            env='dev',
            group='forrest',
            items=items,
            pipeline_settings={
                'lambda': {
                    'vpc_enabled': False,
                },
            },
            region='us-east-1',
            service=service)
    except json.decoder.JSONDecodeError:
        pytest.fail('Bad template: {0}'.format(template_name), pytrace=False)

    assert isinstance(rendered, list)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create IAM Instance Profiles, Roles, Users, and Groups."""
import collections
import logging

import boto3

from ..utils import get_details, get_properties, get_template
from .construct_policy import construct_policy
from .resource_action import resource_action

LOG = logging.getLogger(__name__)


def create_iam_resources(env='dev', app='', **_):
    """Create the IAM Resources for the application.

    Args:
        env (str): Deployment environment/account, i.e. dev, stage, prod.
        app (str): Spinnaker Application name.

    Returns:
        True upon successful completion.
    """
    session = boto3.session.Session(profile_name=env)
    client = session.client('iam')

    app_properties = get_properties(env='pipeline')

    generated = get_details(env=env, app=app)
    generated_iam = generated.iam()
    app_details = collections.namedtuple('AppDetails', generated_iam.keys())
    details = app_details(**generated_iam)

    LOG.debug('Application details: %s', details)

    deployment_type = app_properties['type']
    role_trust_template = get_template(
        'infrastructure/iam/trust/{0}_role.json.j2'.format(deployment_type), formats=generated)

    resource_action(
        client,
        action='create_role',
        log_format='Created Role: %(RoleName)s',
        RoleName=details.role,
        AssumeRolePolicyDocument=role_trust_template)
    resource_action(
        client,
        action='create_instance_profile',
        log_format='Created Instance Profile: %(InstanceProfileName)s',
        InstanceProfileName=details.profile)
    attach_profile_to_role(client, role_name=details.role, profile_name=details.profile)

    iam_policy = construct_policy(app=app, group=details.group, env=env, pipeline_settings=app_properties)
    if iam_policy:
        resource_action(
            client,
            action='put_role_policy',
            log_format='Added IAM Policy: %(PolicyName)s',
            raise_errors=True,
            RoleName=details.role,
            PolicyName=details.policy,
            PolicyDocument=iam_policy)

    resource_action(client, action='create_user', log_format='Created User: %(UserName)s', UserName=details.user)
    resource_action(client, action='create_group', log_format='Created Group: %(GroupName)s', GroupName=details.group)
    resource_action(
        client,
        action='add_user_to_group',
        log_format='Added User to Group: %(UserName)s -> %(GroupName)s',
        GroupName=details.group,
        UserName=details.user)

    return True


def attach_profile_to_role(client, role_name='forrest_unicorn_role', profile_name='forrest_unicorn_profile'):
    """Attach an IAM Instance Profile _profile_name_ to Role _role_name_.

    Args:
        role_name (str): Name of Role.
        profile_name (str): Name of Instance Profile.

    Returns:
        True upon successful completion.
    """
    current_instance_profiles = resource_action(
        client,
        action='list_instance_profiles_for_role',
        log_format='Found Instance Profiles for %(RoleName)s.',
        RoleName=role_name)['InstanceProfiles']

    for profile in current_instance_profiles:
        if profile['InstanceProfileName'] == profile_name:
            LOG.info('Found Instance Profile attached to Role: %s -> %s', profile_name, role_name)
            break
    else:
        for remove_profile in current_instance_profiles:
            resource_action(
                client,
                action='remove_role_from_instance_profile',
                log_format='Removed Instance Profile from Role: '
                '%(InstanceProfileName)s -> %(RoleName)s',
                InstanceProfileName=remove_profile['InstanceProfileName'],
                RoleName=role_name)

        resource_action(
            client,
            action='add_role_to_instance_profile',
            log_format='Added Instance Profile to Role: '
            '%(InstanceProfileName)s -> %(RoleName)s',
            InstanceProfileName=profile_name,
            RoleName=role_name)

    return True



#!/usr/bin/env python
#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Foremast installer"""

import setuptools

with open('requirements.txt', 'rt', encoding="ascii") as reqs_file:
    REQUIREMENTS = reqs_file.readlines()

with open('README.rst', encoding="ascii") as readme_file:
    readme_content = readme_file.read()

if __name__ == "__main__":
    setuptools.setup(
        name='foremast',
        description='Tools for creating infrastructure and Spinnaker Pipelines.',
        long_description=readme_content,
        long_description_content_type='text/x-rst',
        author='Foremast',
        author_email='joelvasallo+foremast@gmail.com',
        packages=setuptools.find_packages(where='src'),
        package_dir={'': 'src'},
        setup_requires=['setuptools_scm'],
        use_scm_version={'local_scheme': 'dirty-tag'},
        install_requires=REQUIREMENTS,
        include_package_data=True,
        keywords="aws gcp redbox gogo infrastructure netflixoss python spinnaker foremast",
        url='https://github.com/foremast/foremast',
        download_url='https://github.com/foremast/foremast',
        platforms=['OS Independent'],
        license='Apache License (2.0)',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        entry_points={
            'console_scripts': [
                'create-app=foremast.app.__main__:main',
                'create-configs=foremast.configs.__main__:main',
                'create-dns=foremast.dns.__main__:main',
                'create-elb=foremast.elb.__main__:main',
                'create-iam=foremast.iam.__main__:main',
                'create-pipeline=foremast.pipeline.__main__:main',
                'create-s3=foremast.s3.__main__:main',
                'create-sg=foremast.securitygroup.__main__:main',
                'destroy-dns=foremast.dns.destroy_dns.__main__:main',
                'destroy-elb=foremast.elb.destroy_elb.__main__:main',
                'destroy-iam=foremast.iam.destroy_iam.__main__:main',
                'destroy-s3=foremast.s3.destroy_s3.__main__:main',
                'destroy-sg=foremast.securitygroup.destroy_sg.__main__:main',
                'full-destroy=foremast.destroyer:main',
                'prepare-app-pipeline=foremast.runner:prepare_app_pipeline',
                'foremast-pipeline=foremast.runner:prepare_app_pipeline',
                'prepare-infrastructure=foremast.runner:prepare_infrastructure',
                'foremast-infrastructure=foremast.runner:prepare_infrastructure',
                'prepare-onetime-pipeline=foremast.runner:prepare_onetime_pipeline',
                'foremast-pipeline-onetime=foremast.runner:prepare_onetime_pipeline',
                'create-scaling-policy=foremast.runner:create_scaling_policy',
                'foremast-scaling-policy=foremast.runner:create_scaling_policy',
                'foremast-scheduled-actions=foremast.runner:create_scheduled_actions',
                'rebuild_pipelines=foremast.runner:rebuild_pipelines',
                'foremast-pipeline-rebuild=foremast.runner:rebuild_pipelines',
                'foremast-deploy-s3app=foremast.runner:deploy_s3app',
                'foremast-promote-s3app=foremast.runner:promote_s3app',
                'slack-notify=foremast.slacknotify.__main__:main',
                'foremast=foremast.__main__:main',
            ]
        }, )



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Deploy artifacts to S3."""
import logging
import os
import subprocess

from ..exceptions import S3ArtifactNotFound
from ..utils import get_details, get_properties

LOG = logging.getLogger(__name__)


class S3Deployment:
    """Handle uploading artifacts to S3 and S3 deployment strategies."""

    def __init__(self, app, env, region, prop_path,
                 artifact_path, artifact_version, artifact_branch, primary_region='us-east-1'):
        """S3 deployment object.

        Args:
            app (str): Application name
            env (str): Environment/Account
            region (str): AWS Region
            prop_path (str): Path of environment property file
            artifact_path (str): Path to tar.gz artifact
            primary_region (str): The primary region for the application.
        """
        self.app_name = app
        self.env = env
        self.region = region
        self.artifact_path = artifact_path
        self.version = artifact_version
        self.git_branch = artifact_branch
        self.properties = get_properties(prop_path, env=self.env, region=self.region)
        self.s3props = self.properties['s3']
        generated = get_details(app=app, env=env, region=region)

        include_region = True
        if self.region == primary_region:
            include_region = False
        if self.s3props.get('shared_bucket_master'):
            self.bucket = generated.shared_s3_app_bucket(include_region=include_region)
            self.s3path = app
        elif self.s3props.get('shared_bucket_target'):
            shared_app = self.s3props['shared_bucket_target']
            newgenerated = get_details(app=shared_app, env=env, region=region)
            self.bucket = newgenerated.shared_s3_app_bucket(include_region=include_region)
            self.s3path = app
        elif self.s3props.get('bucket_name'):
            self.bucket = self.s3props['bucket_name']
            self.s3path = app
        else:
            self.bucket = generated.s3_app_bucket(include_region=include_region)
            self.s3path = self.s3props['path'].lstrip('/')

        self.s3_version_path = ''
        self.s3_latest_path = ''
        self.setup_pathing()

    def setup_pathing(self):
        """Format pathing for S3 deployments."""
        self.s3_version_path = self._path_formatter(self.version)
        self.s3_branch_path = self._path_formatter(self.git_branch)
        self.s3_latest_path = self._path_formatter("LATEST")
        self.s3_canary_path = self._path_formatter("CANARY")
        self.s3_alpha_path = self._path_formatter("ALPHA")
        self.s3_mirror_path = self._path_formatter("MIRROR")

    def _path_formatter(self, suffix):
        """Format the s3 path properly.

        Args:
            suffix (str): suffix to add on to an s3 path

        Returns:
            str: formatted path

        """
        if suffix.lower() == "mirror":
            path_items = [self.bucket, self.s3path]
        else:
            path_items = [self.bucket, self.s3path, suffix]

        path = '/'.join(path_items)
        s3_format = "s3://{}"
        formatted_path = path.replace('//', '/')  # removes configuration errors
        full_path = s3_format.format(formatted_path)
        return full_path

    def upload_artifacts(self):
        """Upload artifacts to S3 and copy to correct path depending on strategy."""
        deploy_strategy = self.properties["deploy_strategy"]

        self._upload_artifacts_to_path(deploy_strategy=deploy_strategy)

        # After uploading to version folder, deploy to general folder such as LATEST, CANARY, etc
        if deploy_strategy == "highlander":
            self._sync_to_path(deploy_strategy, self.s3_latest_path)
        elif deploy_strategy == "branchrelease":
            # syncing to latest as a convenience; i.e. LATEST of branch release
            self._sync_to_path(deploy_strategy, "{}/{}".format(self.s3_branch_path, "LATEST"))
        elif deploy_strategy == "canary":
            self._sync_to_path(deploy_strategy, self.s3_canary_path)
        elif deploy_strategy == "alpha":
            self._sync_to_path(deploy_strategy, self.s3_alpha_path)
        elif deploy_strategy == "mirror":
            pass  # Nothing extra needed for mirror deployments
        else:
            raise NotImplementedError

    def promote_artifacts(self, promote_stage='latest'):
        """Promote artifact version to dest.

        Args:
            promote_stage (string): Stage that is being promoted
        """
        deploy_strategy = self.properties["deploy_strategy"]

        if promote_stage.lower() == 'alpha':
            self._sync_to_path(deploy_strategy, self.s3_canary_path)
        elif promote_stage.lower() == 'canary':
            self._sync_to_path(deploy_strategy, self.s3_latest_path)
        else:
            self._sync_to_path(deploy_strategy, self.s3_latest_path)

    def _get_upload_cmd(self, deploy_strategy):
        """Generate the S3 CLI upload command

        Args:
            deploy_strategy (str): Deploy Strategy defined in configuration

        Returns:
            str: The full CLI command to run.
        """
        if deploy_strategy == "mirror":
            dest_path = self.s3_mirror_path
        elif deploy_strategy == "branchrelease":
            dest_path = "{}/{}".format(self.s3_branch_path, self.version)
        else:
            dest_path = self.s3_version_path

        cmd = 'aws s3 sync {} {} --delete --exact-timestamps --profile {}'.format(self.artifact_path,
                                                                                  dest_path, self.env)
        return cmd

    def _upload_artifacts_to_path(self, deploy_strategy):
        """Recursively upload directory contents to S3.

        Args:
            deploy_strategy (str): Deploy Strategy defined in configuration
        """
        if not os.listdir(self.artifact_path) or not self.artifact_path:
            raise S3ArtifactNotFound

        uploaded = False
        if self.s3props.get("content_metadata"):
            LOG.info("Uploading in multiple parts to set metadata")
            uploaded = self.content_metadata_uploads(deploy_strategy)

        if not uploaded:
            cmd = self._get_upload_cmd(deploy_strategy)
            result = subprocess.run(cmd, check=True, shell=True, stdout=subprocess.PIPE)
            LOG.debug("Upload Command Ouput: %s", result.stdout)

        LOG.info("Uploaded artifacts to %s bucket", self.bucket)

    def content_metadata_uploads(self, deploy_strategy):
        """Finds all specified encoded directories and uploads in multiple parts,
        setting metadata for objects.

        Args:
            deploy_strategy (str): Deploy Strategy defined in configuration

        Returns:
            bool: True if uploaded
        """
        excludes_str = ''
        includes_cmds = []
        cmd_base = self._get_upload_cmd(deploy_strategy)

        for content in self.s3props.get('content_metadata'):
            full_path = os.path.join(self.artifact_path, content['path'])
            if not os.listdir(full_path):
                raise S3ArtifactNotFound

            excludes_str += '--exclude "{}/*" '.format(content['path'])
            include_cmd = '{} --exclude "*", --include "{}/*"'.format(cmd_base, content['path'])
            include_cmd += ' --content-encoding {} --metadata-directive REPLACE'.format(content['content-encoding'])
            includes_cmds.append(include_cmd)

        exclude_cmd = '{} {}'.format(cmd_base, excludes_str)
        result = subprocess.run(exclude_cmd, check=True, shell=True, stdout=subprocess.PIPE)
        LOG.info("Uploaded files without metadata with command: %s", exclude_cmd)
        LOG.debug("Upload Command Output: %s", result.stdout)

        for include_cmd in includes_cmds:
            result = subprocess.run(include_cmd, check=True, shell=True, stdout=subprocess.PIPE)
            LOG.info("Uploaded files with metadata with command: %s", include_cmd)
            LOG.debug("Upload Command Output: %s", result.stdout)

        return True

    def _sync_to_path(self, deploy_strategy, target_path):
        """Copy and sync versioned directory to path in S3.

        Args:
            deploy_strategy (str): Deploy Strategy defined in configuration
            target_path (str): Target S3 path to sync based on deploy strategy
        """

        if deploy_strategy == "branchrelease":
            source_path = "{}/{}".format(self.s3_branch_path, self.version)
        else:
            source_path = self.s3_version_path

        cmd_cp = 'aws s3 cp {} {} --recursive --profile {}'.format(source_path, target_path, self.env)
        # AWS CLI sync does not work as expected bucket to bucket with exact timestamp sync.
        cmd_sync = 'aws s3 sync {} {} --delete --exact-timestamps --profile {}'.format(
            source_path, target_path, self.env)

        cp_result = subprocess.run(cmd_cp, check=True, shell=True, stdout=subprocess.PIPE)
        LOG.debug("Copy to %s before sync output: %s", target_path, cp_result.stdout)
        LOG.info("Copied version %s to %s", self.version, target_path)

        sync_result = subprocess.run(cmd_sync, check=True, shell=True, stdout=subprocess.PIPE)
        LOG.debug("Sync to %s command output: %s", target_path, sync_result.stdout)
        LOG.info("Synced version %s to %s", self.version, target_path)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Destroy any DNS records."""
import json
import logging

import boto3

from ...utils import get_details, get_dns_zone_ids, get_template

LOG = logging.getLogger(__name__)


def destroy_dns(app='', env='dev', **_):
    """Destroy DNS records.

    Args:
        app (str): Spinnaker Application name.
        env (str): Deployment environment.
        regions (str): AWS region.

    Returns:
        bool: True upon successful completion.
    """
    client = boto3.Session(profile_name=env).client('route53')

    generated = get_details(app=app, env=env)
    record = generated.dns_elb()

    zone_ids = get_dns_zone_ids(env=env, facing='external')

    for zone_id in zone_ids:
        record_sets = client.list_resource_record_sets(
            HostedZoneId=zone_id, StartRecordName=record, StartRecordType='CNAME', MaxItems='1')

        for found_record in record_sets['ResourceRecordSets']:
            assert destroy_record(client=client, found_record=found_record, record=record, zone_id=zone_id)

    return True


def destroy_record(client=None, found_record=None, record='', zone_id=''):
    """Destroy an individual DNS record.

    Args:
        client (botocore.client.Route53): Route 53 boto3 client.
        found_record (dict): Route 53 record set::

            {'Name': 'unicorn.forrest.dev.example.com.',
             'ResourceRecords':
             [{'Value':
               'internal-unicornforrest-1777489395.us-east-1.elb.amazonaws.com'
               }],
             'TTL': 60,
             'Type': 'CNAME'}

        record (str): Application DNS record name. e.g.
        zone_id (str): Route 53 Hosted Zone ID, e.g. /hostedzone/ZSVGJWJ979WQD.

    Returns:
        bool: True upon successful completion.
    """
    LOG.debug('Found DNS record: %s', found_record)

    if found_record['Name'].strip('.') == record:
        dns_json = get_template(template_file='destroy/destroy_dns.json.j2', record=json.dumps(found_record))
        dns_dict = json.loads(dns_json)

        client.change_resource_record_sets(HostedZoneId=zone_id, ChangeBatch=dns_dict)
        LOG.info('Destroyed "%s" in %s', found_record['Name'], zone_id)
    else:
        LOG.info('DNS record "%s" missing from %s.', record, zone_id)
        LOG.debug('Found someone else\'s record: %s', found_record['Name'])

    return True



"""IAM Role functions."""
import logging

import boto3

LOG = logging.getLogger(__name__)


def get_role_arn(role_name, env, region):
    """Get role ARN given role name.

    Args:
        role_name (str): Role name to lookup
        env (str): Environment in which to lookup
        region (str): Region

    Returns:
        ARN if role found

    """
    session = boto3.Session(profile_name=env, region_name=region)
    iam_client = session.client('iam')

    LOG.debug('Searching for %s.', role_name)

    role = iam_client.get_role(RoleName=role_name)
    role_arn = role['Role']['Arn']

    LOG.debug("Found role's %s ARN %s", role_name, role_arn)

    return role_arn



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Generic boto3 Resource action caller."""
import logging

from boto3.exceptions import botocore

LOG = logging.getLogger(__name__)


def resource_action(client, action='', log_format='item: %(key)s', raise_errors=False, **kwargs):
    """Call _action_ using boto3 _client_ with _kwargs_.

    This is meant for _action_ methods that will create or implicitly prove a
    given Resource exists. This method will fail silently and log errors by default,
    use the raise_errors flag to change this functionality.  AccessDenied and ExpiredToken
    errors will always cause a raised error. EntityAlreadyExists will never raise an error.

    Args:
        client (botocore.client.IAM): boto3 client object.
        action (str): Client method to call.
        log_format (str): Generic log message format, 'Added' or 'Found' will
            be prepended depending on the scenario.
        raise_errors (bool): If errors should be raised, defaults to False.
            AccessDenied errors will always cause a raised error.
            EntityAlreadyExists will never raise an error.
        **kwargs: Keyword arguments to pass to _action_ method.

    Returns:
        dict: boto3 response.
    """
    result = None
    always_raise_error_codes = ["AccessDenied", "ExpiredToken"]
    try:
        result = getattr(client, action)(**kwargs)
        LOG.info(log_format, kwargs)
    except botocore.exceptions.ClientError as error:
        error_code = error.response['Error']['Code']

        if raise_errors or error_code in always_raise_error_codes:
            LOG.fatal(error)
            raise
        elif error_code == 'EntityAlreadyExists':
            LOG.info(' '.join(('Found', log_format)), kwargs)
        else:
            LOG.fatal(error)

    return result



#   Foremast - Pipeline Tooling
#
#   Copyright 2020 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""GCP IAM Resource management"""

from googleapiclient.errors import HttpError

from ..utils import get_template_object
from ..exceptions import ForemastError
from ..utils.gcp_environment import GcpEnvironment
from . import get_policy, set_policy, modify_policy_remove_member, modify_policy_add_binding
from tryagain import retries
import googleapiclient.discovery
import json
import logging

LOG = logging.getLogger(__name__)


class GcpIamResourceClient:

    def __init__(self, env, app_name, group_name, repo_name, configs):
        self._env = env
        self._credentials = env.get_credentials()
        self._configs = configs
        self._pipeline_type = configs["pipeline"]["type"]
        self._app_name = app_name
        self._group_name = group_name
        self._repo_name = repo_name
        self._services = configs["pipeline"].get('services', dict())

    def create_iam_resources(self):
        """Creates IAM resources including service accounts and IAM role bindings

        Returns:
            Str: the email of the svc account that was created/updated with IAM bindings"""

        # Ensure service account is created
        service_account = self._create_service_account()
        service_account_name = service_account["name"]
        service_account_email = service_account["email"]
        member = "serviceAccount:" + service_account_email
        # Update individual svc account's IAM Policy
        # This is optional and used to grant access TO this svc account, NOT to grant this svc account
        # access to other resources
        self._update_policy_for_service_account(service_account_name)

        # Get the gcp_roles requested in pipeline.json with the full project id as the key
        gcp_roles_by_project = GcpIamResourceClient._get_gcp_roles_by_project(self._services.get('gcp_roles', list()),
                                                                              self._env, self._group_name)

        # For each project in this GCP environment, get the policy
        # remove any references to this svc account, then re-add based on pipeline.json
        # If changes were made locally, update the policy in GCP
        for project in self._env.get_all_projects():
            GcpIamResourceClient._update_policy_for_project(project['projectId'], gcp_roles_by_project, member,
                                                            self._credentials)

        LOG.info("Finished configuring GCP IAM")
        return service_account_email

    def _get_service_account(self):
        """Gets a service account
            Returns:
                dict, Service account with the following keys:
                    name
                    email
                    projectId
                    uniqueId
                    displayName
                    description
            """
        svc_account_name = "projects/{project}/serviceAccounts/{name}@{project}.iam.gserviceaccount.com"\
                           .format(project=self._get_service_account_project(), name=self._app_name)
        service = googleapiclient.discovery.build(
            'iam', 'v1', credentials=self._credentials, cache_discovery=False)
        try:
            response = service.projects().serviceAccounts().get(name=svc_account_name).execute()
            return response
        except HttpError as e:
            if 'status' in e.resp and e.resp['status'] == '404':
                LOG.warning("Did not find existing service account %s", svc_account_name)
                return None
            raise e

    @retries(max_attempts=5, wait=lambda n: 2 ** n, exceptions=HttpError)
    def _create_service_account(self):
        """Creates a service account.
        Returns:
            dict, Created service account with the following keys:
                name
                email
                projectId
                uniqueId
                displayName
                description
        """

        existing_account = self._get_service_account()
        if existing_account:
            LOG.info("GCP service account %s already exists", existing_account['email'])
            return existing_account

        LOG.info("GCP service account for pipeline does not already exist")
        svc_project = self._get_service_account_project()

        service = googleapiclient.discovery.build(
            'iam', 'v1', credentials=self._credentials, cache_discovery=False)
        try:
            app_svc_account = service.projects().serviceAccounts().create(
                name='projects/' + svc_project,
                body={
                    'accountId': self._app_name,
                    'serviceAccount': {
                        'displayName': self._app_name,
                        'description': 'Managed by Foremast'
                    }
                }).execute()

            LOG.info("Created GCP service account with email %s", app_svc_account['email'])
            return app_svc_account
        except HttpError as e:
            # Conflict while updating
            if 'status' in e.resp and e.resp['status'] == '409':
                LOG.warning("GCP returned 409 conflict, this service account may already exist. "
                            "Will retry up to 5 times per project.")
            raise e

    def _get_service_account_project(self):
        """Gets the projectId that should be used when creating the svc account"""
        if self._pipeline_type == "cloudfunction":
            # Override to the same project the function will be deployed to
            cloud_function_project = self._env.get_project(self._configs["pipeline"]["cloudfunction"]["project_name"])
            return cloud_function_project["projectId"]

        # Default is the service account project defined for this env in Foremast config
        return self._env.service_account_project

    @retries(max_attempts=5, wait=lambda n: 2 ** n, exceptions=HttpError)
    def _update_policy_for_service_account(self, resource_name):
        """Updates the IAM policy attached to the given service account
        Args:
            resource_name (str): The service account's full resource name (e.g. projects/../serviceAccounts/..)
        Returns:
            None
        """
        template = get_template_object('infrastructure/iam/gcp-service-account.json.j2')
        rendered_template = template.render(**self._get_jinja_args())
        # If the rendered template is just whitespace, skip the step
        if not rendered_template or rendered_template.isspace():
            LOG.debug("Skipping IAM Policy update for service account '%s' (this is not the same as updating IAM "
                      "bindings on projects)", resource_name)
            return
        # Get/update svc account's IAM Policy
        try:
            service_account_api = googleapiclient.discovery.build(
                'iam', 'v1', credentials=self._credentials, cache_discovery=False).projects().serviceAccounts()
            iam_policy = service_account_api.getIamPolicy(resource=resource_name).execute()
            iam_policy["bindings"] = json.loads(rendered_template)
            body_payload = {
                "policy": iam_policy
            }
            LOG.info("Updating svc account '%s' IAM policy bindings: '%s'", resource_name, rendered_template)
            service_account_api.setIamPolicy(resource=resource_name, body=body_payload).execute()
        except HttpError as e:
            # Conflict while updating
            if 'status' in e.resp and e.resp['status'] == '409':
                LOG.warning("GCP returned 409 conflict, this service account policy was recently updated elsewhere.  "
                            "Will retry up to 5 times per service account policy.")
            raise e

    def _get_jinja_args(self):
        return {
            'app': self._app_name,
            'group': self._group_name,
            'repo': self._repo_name,
            'pipeline_type': self._pipeline_type,
            'env': self._env.name
        }

    @staticmethod
    @retries(max_attempts=5, wait=lambda n: 2 ** n, exceptions=HttpError)
    def _update_policy_for_project(project_id, roles_by_project, member, credentials):
        # Policies in GCP are shared between users/service accounts
        # Attempt to update this project policy, if GCP returns a
        # a 409 conflict we must re-retrieve the policy, re-update it
        # and attempt to save again to save it again
        try:
            LOG.info("Preparing to update IAM policy for project %s", project_id)
            policy = get_policy(credentials, project_id)
            # Remove any references to this svc account from policy
            policy_was_updated = modify_policy_remove_member(policy, member)

            # Update policy
            if project_id in roles_by_project:
                policy_was_updated = True
                # Add all roles in pipeline.json for this project
                for role in roles_by_project[project_id]['roles']:
                    modify_policy_add_binding(policy=policy, role=role, member=member)

            # if the policy was edited, send to Google APIs
            if policy_was_updated:
                LOG.info("IAM Policy for project {} will be updated because it was modified locally"
                         .format(project_id))
                set_policy(credentials=credentials, project_id=project_id, policy=policy)
            else:
                LOG.info("IAM Policy for project {} will not be updated because it was not modified locally"
                         .format(project_id))
        except HttpError as e:
            # Conflict while updating
            if 'status' in e.resp and e.resp['status'] == '409':
                LOG.warning("GCP returned 409 conflict, this policy was recently updated elsewhere.  "
                            "Will retry up to 5 times per project.")
            raise e

    @staticmethod
    def _get_gcp_roles_by_project(gcp_roles: list, env: GcpEnvironment, group_name: str):
        """Gets the full project name for the given services.gcp_roles block

        Returns a dictionary where the key is the full project name, and the value
        is the corresponding entry from services.gcp_roles

        Raises ForemastError if a requested role is not permitted on the given project"""
        roles_by_project = dict()

        for role_definition in gcp_roles:
            project = env.get_project(project_name=role_definition['project_name'])
            # If they are not permitted to request roles on this project, stop Foremast now:
            if not GcpIamResourceClient._check_group_permitted_project(project, group_name):
                raise ForemastError("Group '{}' is not permitted to request roles in "
                                    "project '{}'".format(group_name, project['projectId']))

            # Cleared to request roles in this project
            # Check if the project key already exists, if so merge the roles.  This can happen
            # when Foremast adds a minimum set of roles, and the user also requested their own roles
            project_id = project['projectId']
            if project_id in roles_by_project:
                roles_by_project[project_id]["roles"].extend(role_definition["roles"])
            else:
                roles_by_project[project_id] = role_definition

        return roles_by_project

    @staticmethod
    def _check_group_permitted_project(project, group_name):
        """Returns true if a group is permitted to access this project based on
        the projects FOREMAST_GROUPS=team1,team2 label.  If no label is present
        all groups are permitted."""

        if 'foremast_groups' not in project['labels'] or not project['labels']['foremast_groups'].strip():
            LOG.debug("Project '%s' permits access from all groups", project['projectId'])
            return True

        # foremast_groups defined and not empty
        foremast_groups = project['labels']['foremast_groups'].split('__')  # split on two underscores
        LOG.debug("Project '%s' only supports foremast deployments from groups: '%s'", project['projectId'],
                  foremast_groups)
        for permitted_group in foremast_groups:
            if group_name == permitted_group.strip():
                return True

        # Default to denied
        return False

    @staticmethod
    def list_service_accounts(credentials, project_id=None):
        """Lists all service accounts for the given project.
        Args:
            credentials: GCP API Credentials for a service account
            project_id (str): The project to target
        Returns:
            list: List of service accounts, empty list if none exist
        """
        service = googleapiclient.discovery.build(
            'iam', 'v1', credentials=credentials, cache_discovery=False)

        service_accounts = service.projects().serviceAccounts().list(
            name='projects/' + project_id).execute()

        return service_accounts.get('accounts', list())



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test create_pipeline functionality"""
from unittest import mock

import pytest
from foremast.pipeline import SpinnakerPipeline

TEST_FORMAT_GENERATOR = mock.Mock()
TEST_SETTINGS = {
    'dev': {
        'regions': ['us-east-1'],
        'us-east-1': {
            'app': {
                'app_description': 'Test App Demo application'
            },
            'deploy_strategy': 'highlander',
            'regions': ['us-east-1'],
        }
    },
    'pipeline': {
        'base': 'tomcat8',
        'config_commit': '',
        'deployment': 'spinnaker',
        'documentation': '',
        'env': ['dev'],
        'eureka': True,
        'image': {
            'builder': 'ebs',
            'root_volume_size': 6
        },
        'regions': ['us-east-1', 'us-west-2'],
        'type': 'ec2'
    }
}


@pytest.fixture
@mock.patch('foremast.pipeline.create_pipeline.get_properties')
@mock.patch('foremast.pipeline.create_pipeline.get_details')
@mock.patch('foremast.pipeline.create_pipeline.os')
def spinnaker_pipeline(mock_os, mock_get_details, mock_get_prop):
    """Sets up pipeline fixture object"""
    mock_get_prop.return_value = TEST_SETTINGS
    pipelineObj = SpinnakerPipeline(
        app='appgroup',
        trigger_job='a_group_app', )
    pipelineObj.generated = TEST_FORMAT_GENERATOR
    pipelineObj.app_name = 'appgroup'
    pipelineObj.group_name = 'group'
    return pipelineObj


@mock.patch('foremast.pipeline.create_pipeline.clean_pipelines')
@mock.patch.object(SpinnakerPipeline, 'render_wrapper')
@mock.patch('foremast.pipeline.create_pipeline.get_subnets')
@mock.patch('foremast.pipeline.create_pipeline.construct_pipeline_block')
@mock.patch('foremast.pipeline.create_pipeline.renumerate_stages')
@mock.patch.object(SpinnakerPipeline, 'post_pipeline')
def test_create_pipeline_ec2(mock_post, mock_renumerate, mock_construct, mock_subnets, mock_wrapper, mock_clean,
                             spinnaker_pipeline):
    """test pipeline creation if ec2 pipeline."""
    test_block_data = {
        "env": "dev",
        "generated": TEST_FORMAT_GENERATOR,
        "previous_env": None,
        "region": "us-east-1",
        "settings": spinnaker_pipeline.settings["dev"]["us-east-1"],
        "pipeline_data": spinnaker_pipeline.settings['pipeline'],
        "region_subnets": {
            'us-east-1': ['us-east-1d', 'us-east-1a', 'us-east-1e']
        }
    }
    mock_subnets.return_value = {'dev': {'us-east-1': ['us-east-1d', 'us-east-1a', 'us-east-1e']}}
    mock_construct.return_value = '{"test": "stuff"}'
    mock_wrapper.return_value = {'stages': []}
    created = spinnaker_pipeline.create_pipeline()

    mock_construct.assert_called_with(**test_block_data)
    mock_post.assert_called_with({'stages': ['test']})

    assert created == True



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module to create dynamically generated DNS record in route53"""
import logging

from ..consts import DOMAIN
from ..utils import (find_elb, find_elb_dns_zone_id, get_details, get_dns_zone_ids, get_properties,
                     update_dns_zone_record, update_failover_dns_record)


class SpinnakerDns:
    """Manipulate and create generated DNS record in Route53.

    Args:
        app (str): application name for DNS record
        env (str): Environment/Account for DNS record creation
        region (str): AWS Region for DNS record
        elb_subnet (str): Wether the DNS record is in a public or private zone
        prop_path (str): Path to the generated property files

    Returns:
        str: FQDN of application

    """

    def __init__(self, app=None, env=None, region=None, elb_subnet=None, prop_path=None):
        self.log = logging.getLogger(__name__)

        self.domain = DOMAIN
        self.env = env
        self.region = region
        self.elb_subnet = elb_subnet
        self.app = app

        self.generated = get_details(app, env=self.env, region=self.region)
        self.app_name = self.generated.app_name()

        self.properties = get_properties(properties_file=prop_path, env=self.env, region=self.region)
        self.dns_ttl = self.properties['dns']['ttl']
        self.header = {'content-type': 'application/json'}

    def create_elb_dns(self, regionspecific=False):
        """Create dns entries in route53.

        Args:
            regionspecific (bool): The DNS entry should have region on it
        Returns:
            str: Auto-generated DNS name for the Elastic Load Balancer.

        """
        if regionspecific:
            dns_elb = self.generated.dns()['elb_region']
        else:
            dns_elb = self.generated.dns()['elb']

        dns_elb_aws = find_elb(name=self.app_name, env=self.env, region=self.region)

        zone_ids = get_dns_zone_ids(env=self.env, facing=self.elb_subnet)

        self.log.info('Updating Application URL: %s', dns_elb)

        dns_kwargs = {
            'dns_name': dns_elb,
            'dns_name_aws': dns_elb_aws,
            'dns_ttl': self.dns_ttl,
        }

        for zone_id in zone_ids:
            self.log.debug('zone_id: %s', zone_id)
            update_dns_zone_record(self.env, zone_id, **dns_kwargs)

        return dns_elb

    def create_failover_dns(self, primary_region='us-east-1'):
        """Create dns entries in route53 for multiregion failover setups.

        Args:
            primary_region (str): primary AWS region for failover
        Returns:
            Auto-generated DNS name.
        """
        dns_record = self.generated.dns()['global']
        zone_ids = get_dns_zone_ids(env=self.env, facing=self.elb_subnet)

        elb_dns_aws = find_elb(name=self.app_name, env=self.env, region=self.region)
        elb_dns_zone_id = find_elb_dns_zone_id(name=self.app_name, env=self.env, region=self.region)

        if primary_region in elb_dns_aws:
            failover_state = 'PRIMARY'
        else:
            failover_state = 'SECONDARY'
        self.log.info("%s set as %s record", elb_dns_aws, failover_state)

        self.log.info('Updating Application Failover URL: %s', dns_record)

        dns_kwargs = {
            'dns_name': dns_record,
            'elb_dns_zone_id': elb_dns_zone_id,
            'elb_aws_dns': elb_dns_aws,
            'dns_ttl': self.dns_ttl,
            'failover_state': failover_state,
        }

        for zone_id in zone_ids:
            self.log.debug('zone_id: %s', zone_id)
            update_failover_dns_record(self.env, zone_id, **dns_kwargs)

        return dns_record



"""Test API Gateway functions."""
from unittest import mock

from foremast.awslambda.api_gateway_event.api_gateway_event import APIGateway

TEST_RULES = {'api_name': 1, 'method': 'PUT'}


@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.boto3')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_details')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_env_credential')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_properties')
def test_apigateway(get_properties, get_env_credential, get_details, boto3):
    """Check basic object initialization."""
    test = APIGateway(rules=TEST_RULES)
    assert test


@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.boto3')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_details')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_env_credential')
@mock.patch('foremast.awslambda.api_gateway_event.api_gateway_event.get_properties')
def test_create_resource(get_properties, get_env_credential, get_details, boto3):
    """Check creating API Resource."""
    test = APIGateway(rules=TEST_RULES)
    test.api_id = ''
    test.create_resource()

    test.client.create_resource.assert_called_with(restApiId='', parentId='', pathPart='')



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Package for foremast supporting utilities."""
from .apps import *
from .asg import *
from .banners import *
from .pipelines import *
from .deep_chain_map import DeepChainMap
from .elb import *
from .encoding import *
from .foremast_configs import *
from .generate_filename import *
from .dns import *
from .credentials import *
from .properties import *
from .security_group import *
from .subnets import *
from .vpc import *
from .lookups import *
from .slack import *
from .tasks import *
from .templates import *
from .warn_user import *
from .get_cloudwatch_event_rule import get_cloudwatch_event_rule
from .awslambda import *
from .get_sns_subscriptions import get_sns_subscriptions
from .get_sns_topic_arn import get_sns_topic_arn
from .dynamodb_stream import get_dynamodb_stream_arn
from .roles import *
from .backoff import *



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Create ELBs for Spinnaker Pipelines."""
import json
import logging
from pprint import pformat

import boto3

from ..consts import DEFAULT_ELB_SECURITYGROUPS
from ..utils import get_properties, get_subnets, get_template, get_vpc_id, remove_duplicate_sg, wait_for_task
from .format_listeners import format_listeners
from .splay_health import splay_health

LOG = logging.getLogger(__name__)


class SpinnakerELB:
    """Create ELBs for Spinnaker.

    Args:
        app (str): Application name.
        env (str): Deployment environment.
        prop_path (str): Path to the raw.properties.json.
        region (str): AWS Region.
    """

    def __init__(self, app='', env='', region='', prop_path=''):
        self.app = app
        self.env = env
        self.region = region
        self.properties = get_properties(properties_file=prop_path, env=self.env, region=self.region)

    def make_elb_json(self):
        """Render the JSON template with arguments.

        Returns:
            str: Rendered ELB template.
        """
        env = self.env
        region = self.region
        elb_settings = self.properties['elb']
        LOG.debug('Block ELB Settings:\n%s', pformat(elb_settings))

        health_settings = elb_settings['health']
        elb_subnet_purpose = elb_settings.get('subnet_purpose', 'internal')

        region_subnets = get_subnets(target='elb', purpose=elb_subnet_purpose, env=env, region=region)
        region_subnets.pop("subnet_ids", None)

        # CAVEAT: Setting the ELB to public, you must use a public subnet,
        #         otherwise AWS complains about missing IGW on subnet.

        if elb_subnet_purpose == 'internal':
            is_internal = 'true'
        else:
            is_internal = 'false'

        target = elb_settings.get('target', 'HTTP:80/health')
        health = splay_health(target)

        listeners = format_listeners(elb_settings=elb_settings, env=self.env, region=region)

        idle_timeout = elb_settings.get('idle_timeout', None)
        access_log = elb_settings.get('access_log', {})
        connection_draining_timeout = elb_settings.get('connection_draining_timeout', None)

        security_groups = DEFAULT_ELB_SECURITYGROUPS[env]
        security_groups.append(self.app)
        security_groups.extend(self.properties['security_group']['elb_extras'])
        security_groups = remove_duplicate_sg(security_groups)

        template_kwargs = {
            'access_log': json.dumps(access_log),
            'app_name': self.app,
            'availability_zones': json.dumps(region_subnets),
            'connection_draining_timeout': json.dumps(connection_draining_timeout),
            'env': env,
            'hc_string': target,
            'health_interval': health_settings['interval'],
            'health_path': health.path,
            'health_port': health.port,
            'health_protocol': health.proto,
            'health_timeout': health_settings['timeout'],
            'healthy_threshold': health_settings['threshold'],
            'idle_timeout': json.dumps(idle_timeout),
            'isInternal': is_internal,
            'listeners': json.dumps(listeners),
            'region_zones': json.dumps(region_subnets[region]),
            'region': region,
            'security_groups': json.dumps(security_groups),
            'subnet_type': elb_subnet_purpose,
            'unhealthy_threshold': health_settings['unhealthy_threshold'],
            'vpc_id': get_vpc_id(env, region),
        }

        rendered_template = get_template(template_file='infrastructure/elb_data.json.j2', **template_kwargs)

        return rendered_template

    def create_elb(self):
        """Create or Update the ELB after rendering JSON data from configs.
        Asserts that the ELB task was successful.
        """

        json_data = self.make_elb_json()
        LOG.debug('Block ELB JSON Data:\n%s', pformat(json_data))

        wait_for_task(json_data)

        self.add_listener_policy(json_data)
        self.add_backend_policy(json_data)

        self.configure_attributes(json_data)

    def add_listener_policy(self, json_data):
        """Attaches listerner policies to an ELB

        Args:
            json_data (json): return data from ELB upsert
        """
        env = boto3.session.Session(profile_name=self.env, region_name=self.region)
        elbclient = env.client('elb')

        # create stickiness policy if set in configs
        stickiness = {}
        elb_settings = self.properties['elb']
        if elb_settings.get('ports'):
            ports = elb_settings['ports']
            for listener in ports:
                if listener.get("stickiness"):
                    stickiness = self.add_stickiness()
                    LOG.info('Stickiness Found: %s', stickiness)
                    break

        # Attach policies to created ELB
        for job in json.loads(json_data)['job']:
            for listener in job['listeners']:
                policies = []
                ext_port = listener['externalPort']
                if listener['listenerPolicies']:
                    policies.extend(listener['listenerPolicies'])
                if stickiness.get(ext_port):
                    policies.append(stickiness.get(ext_port))
                if policies:
                    LOG.info('Adding listener policies: %s', policies)
                    elbclient.set_load_balancer_policies_of_listener(
                        LoadBalancerName=self.app, LoadBalancerPort=ext_port, PolicyNames=policies)

    def add_backend_policy(self, json_data):
        """Attaches backend server policies to an ELB

        Args:
            json_data (json): return data from ELB upsert
        """
        env = boto3.session.Session(profile_name=self.env, region_name=self.region)
        elbclient = env.client('elb')

        # Attach backend server policies to created ELB
        for job in json.loads(json_data)['job']:
            for listener in job['listeners']:
                instance_port = listener['internalPort']
                backend_policy_list = listener['backendPolicies']
                if backend_policy_list:
                    LOG.info('Adding backend server policies: %s', backend_policy_list)
                    elbclient.set_load_balancer_policies_for_backend_server(
                        LoadBalancerName=self.app, InstancePort=instance_port, PolicyNames=backend_policy_list)

    def add_stickiness(self):
        """ Adds stickiness policy to created ELB

        Returns:
            dict: A dict of stickiness policies and ports::

                example:
                {
                    80: "$policy_name"
                }
        """
        stickiness_dict = {}
        env = boto3.session.Session(profile_name=self.env, region_name=self.region)
        elbclient = env.client('elb')
        elb_settings = self.properties['elb']
        for listener in elb_settings.get('ports'):
            if listener.get("stickiness"):
                sticky_type = listener['stickiness']['type'].lower()
                externalport = int(listener['loadbalancer'].split(":")[-1])
                policyname_tmp = "{0}-{1}-{2}-{3}"
                if sticky_type == 'app':
                    cookiename = listener['stickiness']['cookie_name']
                    policy_key = cookiename.replace('.', '')
                    policyname = policyname_tmp.format(self.app, sticky_type, externalport, policy_key)
                    elbclient.create_app_cookie_stickiness_policy(
                        LoadBalancerName=self.app, PolicyName=policyname, CookieName=cookiename)
                    stickiness_dict[externalport] = policyname
                elif sticky_type == 'elb':
                    cookie_ttl = listener['stickiness'].get('cookie_ttl', None)
                    policyname = policyname_tmp.format(self.app, sticky_type, externalport, cookie_ttl)
                    if cookie_ttl:
                        elbclient.create_lb_cookie_stickiness_policy(
                            LoadBalancerName=self.app, PolicyName=policyname, CookieExpirationPeriod=cookie_ttl)
                    else:
                        elbclient.create_lb_cookie_stickiness_policy(LoadBalancerName=self.app, PolicyName=policyname)
                    stickiness_dict[externalport] = policyname
        return stickiness_dict

    def configure_attributes(self, json_data):
        """Configure load balancer attributes such as idle timeout, connection draining, etc

        Args:
            json_data (json): return data from ELB upsert
        """
        env = boto3.session.Session(profile_name=self.env, region_name=self.region)
        elbclient = env.client('elb')

        elb_settings = self.properties['elb']
        LOG.debug('Block ELB Settings Pre Configure Load Balancer Attributes:\n%s', pformat(elb_settings))

        # FIXME: Determine why 'job' is not being used
        # pylint: disable=unused-variable
        for job in json.loads(json_data)['job']:
            load_balancer_attributes = {
                'CrossZoneLoadBalancing': {
                    'Enabled': True
                },
                'AccessLog': {
                    'Enabled': False,
                },
                'ConnectionDraining': {
                    'Enabled': False,
                },
                'ConnectionSettings': {
                    'IdleTimeout': 60
                }
            }
            if elb_settings.get('connection_draining_timeout'):
                connection_draining_timeout = int(elb_settings['connection_draining_timeout'])
                LOG.info('Applying Custom Load Balancer Connection Draining Timeout: %d', connection_draining_timeout)
                load_balancer_attributes['ConnectionDraining'] = {
                    'Enabled': True,
                    'Timeout': connection_draining_timeout
                }
            if elb_settings.get('idle_timeout'):
                idle_timeout = int(elb_settings['idle_timeout'])
                LOG.info('Applying Custom Load Balancer Idle Timeout: %d', idle_timeout)
                load_balancer_attributes['ConnectionSettings'] = {'IdleTimeout': idle_timeout}
            if elb_settings.get('access_log'):
                access_log_bucket_name = elb_settings['access_log']['bucket_name']
                access_log_bucket_prefix = elb_settings['access_log']['bucket_prefix']
                access_log_emit_interval = int(elb_settings['access_log']['emit_interval'])
                LOG.info('Applying Custom Load Balancer Access Log: %s/%s every %d minutes', access_log_bucket_name,
                         access_log_bucket_prefix, access_log_emit_interval)
                load_balancer_attributes['AccessLog'] = {
                    'Enabled': True,
                    'S3BucketName': access_log_bucket_name,
                    'EmitInterval': access_log_emit_interval,
                    'S3BucketPrefix': access_log_bucket_prefix
                }

            LOG.info('Applying Load Balancer Attributes')
            LOG.debug('Load Balancer Attributes:\n%s', pformat(load_balancer_attributes))
            elbclient.modify_load_balancer_attributes(
                LoadBalancerName=self.app, LoadBalancerAttributes=load_balancer_attributes)



#   Foremast - Pipeline Tooling
#
#   Copyright 2019 Redbox Automated Retail, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Functions and variables that can be exposed to Jinja2 templates"""

from ..utils.kayenta import get_canary_id
from ..utils.pipelines import get_pipeline_id, generate_predictable_pipeline_id


def get_jinja_functions():
    """Gets a dictionary of functions that can be exposed to Jinja templates"""
    functions = dict()
    functions[get_canary_id.__name__] = get_canary_id
    functions[get_pipeline_id.__name__] = get_pipeline_id
    functions[raise_exception.__name__] = raise_exception
    functions[generate_predictable_pipeline_id.__name__] = generate_predictable_pipeline_id

    return functions


def get_jinja_variables(pipeline):
    """Gets a dictionary of variables from a SpinnakerPipeline that can be exposed to Jinja templates"""
    variables = dict()
    # Deprecated variables: Use the app block instead
    variables["trigger_job"] = pipeline.trigger_job
    variables["group_name"] = pipeline.group_name
    variables["app_name"] = pipeline.app_name
    variables["repo_name"] = pipeline.repo_name
    # Deprecated end

    email = pipeline.settings['pipeline']['notifications']['email']
    slack = pipeline.settings['pipeline']['notifications']['slack']
    deploy_type = pipeline.settings['pipeline']['type']

    # Replaces top level variables above which are deprecated
    # app block matches non-manual pipeline types like ec2, lambda, etc for consistency
    variables["data"] = {
        'app': {
            'appname': pipeline.app_name,
            'group_name': pipeline.group_name,
            'repo_name': pipeline.repo_name,
            'deploy_type': deploy_type,
            'environment': 'packaging',
            'triggerjob': pipeline.trigger_job,
            'email': email,
            'slack': slack,
            'pipeline': pipeline.settings['pipeline']
        }
    }

    return variables


def raise_exception(message):
    """A function for raising an exception from inside a Jinja Template"""
    raise Exception(message)



#   Foremast - Pipeline Tooling
#
#   Copyright 2018 Gogo, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test IAM Policies for correctness."""
import copy
import json
from unittest import mock

import pytest

from foremast.iam.construct_policy import construct_policy
from foremast.utils.templates import get_template


@pytest.fixture
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def get_base_settings():
    return json.loads(get_template(template_file='configs/pipeline.json.j2'))


@mock.patch('foremast.utils.credentials.gate_request')
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def test_iam_construct_policy(gate_request, get_base_settings):
    """Check general assemblage."""
    settings = get_base_settings

    policy_json = construct_policy(pipeline_settings=settings)
    # checking empty policy
    assert policy_json is None

    settings.update({'services': {'s3': True}})
    policy_json = construct_policy(app='unicornforrest', env='stage', group='forrest', pipeline_settings=settings)

    # checking s3 policy
    assert type(json.loads(policy_json)) == dict

    # TODO: Test other services besides S3
    settings.update({'services': {'dynamodb': ['coreforrest', 'edgeforrest', 'attendantdevops']}})
    policy_json = construct_policy(pipeline_settings=settings)
    policy = json.loads(policy_json)


@mock.patch('foremast.utils.credentials.gate_request')
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def test_construct_cloudwatchlogs(gate_request, get_base_settings):
    """Check Lambda Policy."""
    pipeline_settings = get_base_settings
    pipeline_settings.update({'services': {'cloudwatchlogs': True}, 'type': 'lambda'})

    policy_json = construct_policy(
        app='unicornforrest', env='dev', group='forrest', pipeline_settings=pipeline_settings)
    policy = json.loads(policy_json)

    statements = list(statement for statement in policy['Statement'] if statement['Sid'] == 'LambdaCloudWatchLogs')
    assert len(statements) == 1

    statement = statements[0]
    assert statement['Effect'] == 'Allow'
    assert len(statement['Action']) == 3
    assert all(action.startswith('logs:') for action in statement['Action'])


@mock.patch('foremast.utils.credentials.gate_request')
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def test_construct_s3(gate_request, get_base_settings):
    """Check S3 Policy."""
    pipeline_settings = get_base_settings
    pipeline_settings.update({'services': {'s3': True}})

    construct_policy_kwargs = {
        'app': 'unicornforrest',
        'env': 'dev',
        'group': 'forrest',
        'pipeline_settings': pipeline_settings
    }

    policy_json = construct_policy(**construct_policy_kwargs)
    policy = json.loads(policy_json)
    assert len(policy['Statement']) == 2

    allow_list_policy, allow_edit_policy = policy['Statement']

    assert len(allow_list_policy['Action']) == 3
    assert 's3:ListBucket' in allow_list_policy['Action']
    assert len(allow_list_policy['Resource']) == 0

    assert len(allow_edit_policy['Action']) == 5
    assert all(('s3:{0}Object'.format(action) in allow_edit_policy['Action'] for action in ('Delete', 'Get', 'Put')))
    assert len(allow_edit_policy['Resource']) == 0


@mock.patch('foremast.utils.credentials.gate_request')
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def test_construct_s3_buckets(gate_request, get_base_settings):
    """Check S3 Policy with multiple Buckets listed."""
    pipeline_settings = get_base_settings
    pipeline_settings.update({'services': {'s3': ['Bucket1', 'Bucket2']}})

    construct_policy_kwargs = {
        'app': 'unicornforrest',
        'env': 'dev',
        'group': 'forrest',
        'pipeline_settings': pipeline_settings
    }

    policy_json = construct_policy(**construct_policy_kwargs)
    policy = json.loads(policy_json)
    print(policy)
    assert len(policy['Statement']) == 2

    allow_list_policy, allow_edit_policy = policy['Statement']

    assert len(allow_list_policy['Resource']) == 2

    assert len(allow_edit_policy['Resource']) == 2


@mock.patch('foremast.utils.credentials.gate_request')
@mock.patch('foremast.utils.templates.TEMPLATES_PATH', None)
def test_construct_sdb_domains(gate_request, get_base_settings):
    """Check SimpleDB Policy with multiple Domains listed."""
    pipeline_settings = get_base_settings
    pipeline_settings.update({'services': {'sdb': ['Domain1', 'Domain2']}})

    construct_policy_kwargs = {
        'app': 'unicornforrest',
        'env': 'dev',
        'group': 'forrest',
        'pipeline_settings': pipeline_settings
    }

    policy_json = construct_policy(**construct_policy_kwargs)
    policy = json.loads(policy_json)
    assert len(policy['Statement']) == 1
    assert len(policy['Statement'][0]['Resource']) == 2
    assert policy['Statement'][0]['Resource'][0].endswith('Domain1')
    assert policy['Statement'][0]['Resource'][1].endswith('Domain2')
